import asyncio
import warnings
from datetime import datetime
from functools import reduce
from typing import Dict, List, Optional, Tuple

import httpx
import pandas as pd

from CurvyCUSIPs.DataFetcher.base import DataFetcherBase
from CurvyCUSIPs.utils.ust_utils import get_isin_from_cusip

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)


class BondSupermartDataFetcher(DataFetcherBase):
    def __init__(
        self,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        super().__init__(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

    async def _fetch_cusip_timeseries_bondsupermart(
        self,
        client: httpx.AsyncClient,
        cusip: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
        uid: Optional[str | int] = None,
    ):
        isin = get_isin_from_cusip(cusip)
        headers = {
            "authority": "www.bondsupermart.com",
            "method": "GET",
            "path": f"/main/ws/v3/bond-info/bond-factsheet-chart/US{isin}",
            "scheme": "https",
            "accept": "application/json, text/plain, */*",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9",
            "cache-control": "no-cache",
            "dnt": "1",
            "pragma": "no-cache",
            "priority": "u=1, i",
            "referer": f"https://www.bondsupermart.com/bsm/bond-factsheet/US{isin}",
            "sec-ch-ua": '"Chromium";v="128", "Not;A=Brand";v="24", "Google Chrome";v="128"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36",
        }
        cols_to_return = ["Date", "bid_price", "bid_yield", "ask_price", "ask_yield"]
        retries = 0
        try:
            while retries < max_retries:
                try:
                    url = f"https://www.bondsupermart.com/main/ws/v3/bond-info/bond-factsheet-chart/{isin}"
                    response = await client.get(url, headers=headers)
                    response.raise_for_status()
                    response_json = response.json()

                    if not response_json["yieldChartMap"] or not response_json["priceChartMap"]:
                        raise ValueError("Data is None")

                    bid_yield_df = pd.DataFrame(
                        [
                            {"Date": ts_yield_list[0], "bid_yield": ts_yield_list[1]}
                            for ts_yield_list in response_json["yieldChartMap"]["SINCE_INCEPTION"][0]["data"]
                        ]
                    )
                    ask_yield_df = pd.DataFrame(
                        [
                            {"Date": ts_yield_list[0], "ask_yield": ts_yield_list[1]}
                            for ts_yield_list in response_json["yieldChartMap"]["SINCE_INCEPTION"][1]["data"]
                        ]
                    )
                    bid_price_df = pd.DataFrame(
                        [
                            {"Date": ts_price_list[0], "bid_price": ts_price_list[1]}
                            for ts_price_list in response_json["priceChartMap"]["SINCE_INCEPTION"][0]["data"]
                        ]
                    )
                    ask_price_df = pd.DataFrame(
                        [
                            {"Date": ts_price_list[0], "ask_price": ts_price_list[1]}
                            for ts_price_list in response_json["priceChartMap"]["SINCE_INCEPTION"][1]["data"]
                        ]
                    )

                    merged_df = reduce(
                        lambda left, right: pd.merge(left, right, on="Date", how="outer"),
                        [bid_price_df, bid_yield_df, ask_price_df, ask_yield_df],
                    )
                    merged_df["Date"] = pd.to_datetime(merged_df["Date"], unit="ms", errors="coerce")
                    merged_df["bid_yield"] = pd.to_numeric(merged_df["bid_yield"], errors="coerce")
                    merged_df["ask_yield"] = pd.to_numeric(merged_df["ask_yield"], errors="coerce")
                    merged_df["mid_yield"] = (merged_df["bid_yield"] + merged_df["ask_yield"]) / 2

                    merged_df["bid_price"] = pd.to_numeric(merged_df["bid_price"], errors="coerce")
                    merged_df["ask_price"] = pd.to_numeric(merged_df["ask_price"], errors="coerce")
                    merged_df["mid_price"] = (merged_df["bid_price"] + merged_df["ask_price"]) / 2

                    if start_date:
                        merged_df = merged_df[merged_df["Date"].dt.date >= start_date.date()]
                    if end_date:
                        merged_df = merged_df[merged_df["Date"].dt.date <= end_date.date()]

                    if uid:
                        return cusip, merged_df, uid
                    return cusip, merged_df

                except httpx.HTTPStatusError as e:
                    self._logger.error(f"Bondsupermart.com - Bad Status: {response.status_code}")
                    if response.status_code == 404:
                        if uid:
                            return cusip, pd.DataFrame(columns=cols_to_return), uid
                        return cusip, pd.DataFrame(columns=cols_to_return)

                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST Timeseries GitHub - Throttled for {cusip}. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

                except Exception as e:
                    self._logger.error(f"Bondsupermart.com - Error: {str(e)}")
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"Bondsupermart.com - Throttled for {cusip}. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

            raise ValueError(f"Bondsupermart  - Max retries exceeded for {cusip}")

        except Exception as e:
            self._logger.error(e)
            if uid:
                return cusip, pd.DataFrame(columns=cols_to_return), uid
            return cusip, pd.DataFrame(columns=cols_to_return)

    async def _fetch_cusip_timeseries_bondsupermart_with_semaphore(self, semaphore, *args, **kwargs):
        async with semaphore:
            return await self._fetch_cusip_timeseries_bondsupermart(*args, **kwargs)

    def bondsupermart_timeseries_api(
        self,
        cusips: List[str],
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        max_concurrent_tasks: int = 64,
    ):
        async def build_tasks(
            client: httpx.AsyncClient,
            cusips: List[str],
            start_date: datetime,
            end_date: datetime,
        ):
            semaphore = asyncio.Semaphore(max_concurrent_tasks)
            tasks = [
                self._fetch_cusip_timeseries_bondsupermart_with_semaphore(
                    semaphore=semaphore,
                    client=client,
                    cusip=cusip,
                    start_date=start_date,
                    end_date=end_date,
                )
                for cusip in cusips
            ]
            return await asyncio.gather(*tasks)

        async def run_fetch_all(
            cusips: List[str],
            start_date: datetime,
            end_date: datetime,
        ):
            async with httpx.AsyncClient(proxy=self._proxies["https"]) as client:
                all_data = await build_tasks(
                    client=client,
                    cusips=cusips,
                    start_date=start_date,
                    end_date=end_date,
                )
                return all_data

        dfs: List[Tuple[str, pd.DataFrame]] = asyncio.run(
            run_fetch_all(
                cusips=cusips,
                start_date=start_date,
                end_date=end_date,
            )
        )
        return dict(dfs)



import logging
from typing import Dict, Optional


class DataFetcherBase:
    _global_timeout: int = 10
    _proxies: Dict[str, str] = {"http": None, "https": None}
    _httpx_proxies: Dict[str, str] = {"http://": None, "https://": None}

    _logger = logging.getLogger(__name__)
    _debug_verbose: bool = False
    _error_verbose: bool = False
    _info_verbose: bool = False

    def __init__(
        self,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        self._global_timeout = global_timeout
        self._proxies = proxies if proxies else {"http": None, "https": None}
        self._httpx_proxies["http://"] = self._proxies["http"]
        self._httpx_proxies["https://"] = self._proxies["https"]

        self._debug_verbose = debug_verbose
        self._error_verbose = error_verbose
        self._info_verbose = info_verbose
        self._no_logs_plz = not debug_verbose and not error_verbose and not info_verbose

        self._setup_logger()

    def _setup_logger(self):
        if not self._logger.hasHandlers():
            handler = logging.StreamHandler()
            handler.setFormatter(
                logging.Formatter(
                    "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
                )
            )
            self._logger.addHandler(handler)

        if self._debug_verbose:
            self._logger.setLevel(logging.DEBUG)
        elif self._info_verbose:
            self._logger.setLevel(logging.INFO)
        elif self._error_verbose:
            self._logger.setLevel(logging.ERROR)
        else:
            self._logger.setLevel(logging.WARNING)

        if self._debug_verbose or self._info_verbose or self._error_verbose:
            self._logger.setLevel(logging.DEBUG)

        if self._no_logs_plz:
            self._logger.disabled = True
            self._logger.propagate = False



from types import SimpleNamespace
from typing import List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import seaborn as sns
import statsmodels.api as sm
from scipy import stats
from scipy.odr import ODR, Model, RealData, Data
from scipy.stats import tstd, zscore, linregress

sns.set_style("whitegrid", {"grid.linestyle": "--"})

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)


def run_basic_linear_regression(
    x_series: pd.Series,
    y_series: pd.Series,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    title: Optional[str] = None,
):
    if not x_label:
        x_label = x_series.name
    if not y_label:
        y_label = y_series.name

    Y = y_series
    X = x_series
    X = sm.add_constant(X)
    model = sm.OLS(Y, X)
    results = model.fit()
    print(results.summary())

    intercept = results.params[0]
    slope = results.params[1]
    r_squared = results.rsquared

    plt.figure(figsize=(20, 10))
    plt.scatter(x_series, y_series)

    regression_line = intercept + slope * x_series
    plt.plot(x_series, regression_line, color="red")

    plt.ylabel(x_label)
    plt.xlabel(y_label)
    plt.title(title or f"{y_label} Regressed on {x_label}")
    equation_text = f"y = {intercept:.3f} + {slope:.3f}x\nR² = {r_squared:.3f}\nSE = {results.bse["const"]:.3f}"
    plt.plot([], [], " ", label=f"{equation_text}")

    plt.legend(fontsize="x-large")
    plt.grid(True)
    plt.show()

    return results


def run_basic_linear_regression_df(
    df: pd.DataFrame,
    x_col: str,
    y_col: str,
    title: Optional[str] = None,
    plot_most_recent: Optional[bool] = False,
    date_color_bar: Optional[bool] = False,
    on_diff: Optional[bool] = False,
    run_tls: Optional[bool] = False,
    run_gls: Optional[bool] = False,
    run_wls: Optional[bool | npt.ArrayLike] = None,
):
    if x_col not in df.columns or y_col not in df.columns:
        raise Exception(f"{x_col} or {y_col} not in df cols")

    df = df[["Date"] + [x_col, y_col]].copy()
    if on_diff:
        date_col = df["Date"]
        df = df[[x_col, y_col]].diff()
        df["Date"] = date_col
    df = df.dropna()

    y = df[y_col]
    X = df[x_col]
    X = sm.add_constant(X)

    regression_type = "OLS"

    if run_tls:
        regression_type = "TLS"

        def _linear_f_no_constant(beta: np.array, x_variable: np.array) -> np.array:
            _, b = beta[0], beta[1:]
            b.shape = (b.shape[0], 1)
            return (x_variable * b).sum(axis=0)

        linear = Model(_linear_f_no_constant)
        mydata = RealData(X.T, y)
        myodr = ODR(mydata, linear, beta0=np.ones(X.shape[1] + 1))
        out = myodr.run()
        print(out.pprint())

        hedge_ratios = out.beta[1:]
        residuals = y - (X * hedge_ratios).sum(axis=1)

        x = df[x_col].to_numpy().flatten()
        y = df[y_col].to_numpy().flatten()

        intercept = out.beta[1]
        slope = out.beta[-1]
        sd_intercept = out.sd_beta[1]
        sd_slope = out.sd_beta[0]
        df_resid = len(x) - 2
        results = SimpleNamespace(
            params=pd.Series([intercept, slope], index=["const", x_col]),
            bse=pd.Series([sd_intercept, sd_slope], index=["const", x_col]),
            pvalues=pd.Series(
                [2 * (1 - stats.t.cdf(np.abs(intercept / sd_intercept), df=df_resid)), 2 * (1 - stats.t.cdf(np.abs(slope / sd_slope), df=df_resid))],
                index=["const", x_col],
            ),
            rsquared=1 - np.sum((y - (slope * x + intercept)) ** 2) / np.sum((y - np.mean(y)) ** 2),
            resid=residuals,
            model=SimpleNamespace(**{"endog_names": y_col, "exog_names": ["", x_col]}),
        )

        intercept = results.params[0]
        slope = results.params[1]
        r_squared = results.rsquared
        p_value = results.pvalues[1] if len(results.pvalues) > 1 else None
        slope_name = results.params.drop("const").index[0]
        se_intercept = results.bse[0]
        se_slope = results.bse[1]

    else:
        if run_gls:
            regression_type = "GLS"
            model = sm.GLS(y, X)
        elif run_wls:
            regression_type = "WLS"
            model = sm.WLS(y, X, weights=run_wls)
        else:
            model = sm.OLS(y, X)

        results = model.fit()
        print(results.summary())

        intercept = results.params[0]
        slope = results.params[1]
        r_squared = results.rsquared
        p_value = results.pvalues[1] if len(results.pvalues) > 1 else None
        slope_name = results.params.drop("const").index[0]
        se_intercept = results.bse[0]
        se_slope = results.bse[1]

    plt.figure(figsize=(20, 10))

    if date_color_bar:
        df["date_numeric"] = (df["Date"] - df["Date"].min()).dt.total_seconds()
        scatter = plt.scatter(df[x_col], df[y_col], c=df["date_numeric"], cmap="viridis")
        cbar = plt.colorbar(scatter)
        cbar.set_label("Date")
        cbar_ticks = np.linspace(df["date_numeric"].min(), df["date_numeric"].max(), num=10)
        cbar.set_ticks(cbar_ticks)
        cbar.set_ticklabels(pd.to_datetime(cbar_ticks, unit="s", origin=df["Date"].min()).strftime("%Y-%m-%d"))
    else:
        plt.scatter(df[x_col], df[y_col])

    if plot_most_recent:
        most_recent = df["Date"].iloc[-1]
        plt.scatter(
            df[x_col].iloc[-1],
            df[y_col].iloc[-1],
            color="orange",
            s=100,
            label=f"Most Recent: {most_recent}",
        )

    regression_line = intercept + slope * df[x_col]
    plt.plot(df[x_col], regression_line, color="red")
    plt.xlabel(x_col if not on_diff else f"Δ{x_col}")
    plt.ylabel(y_col if not on_diff else f"Δ{y_col}")
    plt.title(
        title or f"{regression_type} - {y_col} Regressed on {x_col}" if not on_diff else f"{regression_type} - Δ{y_col} Regressed on Δ{x_col}",
        fontdict={"fontsize": "x-large"},
    )
    equation_text = (
        f"y = {intercept:.3f} + {slope:.3f}*{slope_name}\n"
        f"R² = {r_squared:.3f}\n"
        f"SE (Intercept) = {se_intercept:.3f}, SE ({slope_name}) = {se_slope:.3f}\n"
        f"p-value ({slope_name}) = {p_value:.3e}"
    )
    plt.plot([], [], " ", label=f"{equation_text}")
    plt.legend(fontsize="x-large")
    plt.grid(True)
    plt.show()

    return results


def run_multiple_linear_regression_df(
    df: pd.DataFrame,
    x_cols: List[str],
    y_col: str,
    title: Optional[str] = None,
    verbose=False,
    on_diff: Optional[bool] = False,
):
    df = df[["Date"] + x_cols + [y_col]].copy()
    if on_diff:
        date_col = df["Date"]
        df = df[x_cols + [y_col]].diff()
        df["Date"] = date_col
    df = df.dropna()

    if verbose:
        print(df)

    for col in x_cols + [y_col]:
        if col not in df.columns:
            raise Exception(f"{col} not in df columns")

    Y = df[y_col]
    X = df[x_cols]
    X = sm.add_constant(X)

    model = sm.OLS(Y, X)
    results = model.fit()
    print(results.summary())

    intercept = results.params["const"]
    slopes = results.params.drop("const")
    r_squared = results.rsquared
    adj_r_squared = results.rsquared_adj

    Y_pred = results.fittedvalues
    residuals = results.resid

    plt.figure(figsize=(20, 10))
    plt.scatter(Y_pred, Y)
    plt.xlabel("Predicted Values")
    plt.ylabel("Actual Values")
    plt.title("Actual vs Predicted")
    plt.grid(True)

    equation_text = f"y = {intercept:.3f}"
    for col in slopes.index:
        coef = slopes[col]
        equation_text += f" + {coef:.3f}*{col}"
    equation_text += f"\nR² = {r_squared:.3f}\nAdjusted R² = {adj_r_squared:.3f}"
    plt.plot([], [], " ", label=equation_text)
    plt.legend(fontsize="x-large")
    plt.show()

    # Plot Residuals vs Predicted
    plt.figure(figsize=(20, 10))
    plt.scatter(Y_pred, residuals)
    plt.axhline(y=0, color="red", linestyle="--")
    plt.xlabel("Predicted Values")
    plt.ylabel("Residuals")
    plt.title(title or "Residuals vs Predicted")
    plt.grid(True)
    plt.show()

    return df, results


def plot_residuals_timeseries(
    df: pd.DataFrame,
    results: sm.regression.linear_model.RegressionResultsWrapper,
    date_col: str = "Date",
    plot_zscores: Optional[bool] = False,
    stds: Optional[List[int]] = None,
    is_on_diff: Optional[bool] = False,
    rolling_stds: Optional[List[Tuple[int, int]]] = None,
):
    residuals = results.resid
    zscores = zscore(residuals)

    if date_col not in df.columns:
        raise Exception(f"{date_col} not in df columns")

    if is_on_diff:
        df = df.copy()
        df = df.iloc[1:]

    r_squared = results.rsquared
    intercept = results.params[0]
    slope = results.params[1]
    p_value = results.pvalues[1] if len(results.pvalues) > 1 else None
    dependent_variable = results.model.endog_names
    independent_variables = results.model.exog_names[1]
    slope_name = results.params.drop("const").index[0]

    if p_value is not None:
        title = f"Residuals of {dependent_variable} Regressed on {independent_variables} Over Time\n"
    else:
        title = f"Residuals of {dependent_variable} Regressed on {independent_variables} Over Time\n"

    plt.figure(figsize=(20, 10))
    plt.plot(df[date_col], residuals if not plot_zscores else zscores, linestyle="-", color="blue")
    plt.axhline(y=0, color="red", linestyle="--")
    equation_text = f"y = {intercept:.3f} + {slope:.3f}*{slope_name}\nR² = {r_squared:.3f}\nSE = {results.bse["const"]:.3f}"
    for var in results.params.drop("const").index:
        p_val = results.pvalues[var]
        equation_text += f"\np-value ({var}) = {p_val:.3e}"
    plt.plot([], [], " ", label=f"{equation_text}")

    if stds:
        resid_std = tstd(residuals)
        resid_mean = np.mean(residuals)
        plt.axhline(resid_mean, linestyle="--", color="red", label=f"Resid Mean: {resid_mean}")
        for std in stds:
            curr = plt.axhline(resid_mean + resid_std * std, linestyle="--", label=f"+/- {std} STD")
            plt.axhline(resid_mean + resid_std * -1 * std, linestyle="--", color=curr.get_color())

    if rolling_stds:
        for std, window in rolling_stds:
            rolling_resid_std = pd.Series(residuals).rolling(window).std()
            curr = plt.plot(df[date_col], rolling_resid_std * std, linestyle="--", label=f"+/- {std} {window}d Rolling STD")
            plt.plot(df[date_col], -rolling_resid_std * std, linestyle="--", color=curr[0].get_color())

    plt.legend(fontsize="x-large")
    plt.xlabel("Date")
    plt.ylabel("Residuals" if not plot_zscores else "Z-Scores")
    plt.title(title + ", Z-Scroes" if plot_zscores else title, fontdict={"fontsize": "x-large"})
    plt.grid(True)
    plt.show()


def run_rolling_regression_df(df: pd.DataFrame, x_col: str, y_col: str, window: int, title: Optional[str] = None):
    df = df.copy()

    if x_col not in df.columns or y_col not in df.columns:
        raise Exception(f"{x_col} or {y_col} not in df cols")

    def calculate_r_squared(x, y):
        Y = y
        X = x
        X = sm.add_constant(X)
        model = sm.OLS(Y, X)
        results = model.fit()
        r_squared = results.rsquared
        return r_squared

    rolling_r_squared = []
    for rolling_df in df[[x_col, y_col]].rolling(window=window):
        if len(rolling_df.index) < window:
            rolling_r_squared.append(np.nan)
        else:
            r_squared = calculate_r_squared(rolling_df[x_col], rolling_df[y_col])
            rolling_r_squared.append(r_squared)

    plt.figure(figsize=(20, 10))
    plt.plot(df["Date"], rolling_r_squared, label=f"Rolling R-squared (window={window})")

    most_recent = df["Date"].iloc[-1]
    plt.scatter(most_recent, rolling_r_squared[-1], color="purple", s=100, label=f"Most Recent: {most_recent}, R² = {rolling_r_squared[-1]}")

    plt.xlabel("Date")
    plt.ylabel("R-squared")
    plt.title(title or f"Rolling R-squared: {y_col} Regressed on {x_col}", fontdict={"fontsize": "x-large"})

    mean_r_squared = np.nanmean(rolling_r_squared)
    std_r_squared = np.nanstd(rolling_r_squared)
    stats_text = f"Mean R² = {mean_r_squared:.3f}\nStd R² = {std_r_squared:.3f}"
    plt.plot([], [], " ", label=stats_text)

    plt.legend(fontsize="x-large")
    plt.grid(True)
    plt.show()

    return rolling_r_squared


# Orthogonal Distance Regression
"""
https://docs.scipy.org/doc/scipy/reference/odr.html
https://www.mechanicalkern.com/static/odr_ams.pdf
https://www.reddit.com/r/mathematics/comments/12qkes4/confused_btw_orthogonal_linear_regressionolr/
"""
def run_odr(df: pd.DataFrame, x_cols: List[str], y_col: str, x_errs: Optional[npt.ArrayLike] = None, y_errs: Optional[npt.ArrayLike] = None):
    def orthoregress(
        x: pd.Series | npt.ArrayLike, y: pd.Series | npt.ArrayLike, x_errs: Optional[npt.ArrayLike] = None, y_errs: Optional[npt.ArrayLike] = None
    ):
        # calc weights (inverse variances)
        wd = None
        we = None
        if x_errs is not None:
            wd = 1.0 / np.square(x_errs)
        if y_errs is not None:
            we = 1.0 / np.square(y_errs)

        def f(p, x):
            return (p[0] * x) + p[1]

        od = ODR(Data(x, y, wd=wd, we=we), Model(f), beta0=linregress(x, y)[0:2])
        out = od.run()
        return out

    def orthoregress_multilinear(
        X: pd.DataFrame | pd.Series | npt.ArrayLike,
        y: pd.Series | npt.ArrayLike,
        x_errs: Optional[npt.ArrayLike] = None,
        y_errs: Optional[npt.ArrayLike] = None,
    ):
        # calc weights (inverse variances)
        wd = None
        we = None
        if x_errs is not None:
            x_errs = np.asarray(x_errs)
            wd = 1.0 / np.square(x_errs.T)  # transpose to match ODR shape
        if y_errs is not None:
            we = 1.0 / np.square(y_errs)

        def multilinear_f(p, x):
            return np.dot(p[:-1], x) + p[-1]

        X = np.asarray(X)
        y = np.asarray(y)
        X_odr = X.T
        y_flat = y.flatten()
        X_with_intercept = np.column_stack((X, np.ones(X.shape[0])))
        beta_init, _, _, _ = np.linalg.lstsq(X_with_intercept, y_flat, rcond=None)
        beta0 = np.append(beta_init[:-1], beta_init[-1])
        model = Model(multilinear_f)
        data = Data(X_odr, y, wd=wd, we=we)
        odr_instance = ODR(data, model, beta0=beta0)
        output = odr_instance.run()
        return output

    if len(x_cols) > 1:
        out = orthoregress_multilinear(df[x_cols], df[y_col], x_errs, y_errs)
    else:
        out = orthoregress(df[x_cols[0]], df[y_col], x_errs, y_errs)
    out.beta = np.roll(out.beta, 1)
    return out



import asyncio
import warnings
from datetime import datetime
from typing import Dict, List, Optional, Tuple

import httpx
import pandas as pd
import requests

from CurvyCUSIPs.DataFetcher.base import DataFetcherBase

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)


class FedInvestDataFetcher(DataFetcherBase):
    def __init__(
        self,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        super().__init__(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

    # scrapes prices directly from FedInvest
    async def _fetch_cusip_prices_fedinvest(
        self,
        client: httpx.AsyncClient,
        date: datetime,
        cusips: List[str],
        uid: Optional[int | str],
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
    ):
        payload = {
            "priceDate.month": date.month,
            "priceDate.day": date.day,
            "priceDate.year": date.year,
            "submit": "Show Prices",
        }
        headers = {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "Accept-Encoding": "gzip, deflate, br, zstd",
            "Accept-Language": "en-US,en;q=0.9",
            "Cache-Control": "max-age=0",
            "Connection": "keep-alive",
            # "Content-Length": "100",
            "Content-Type": "application/x-www-form-urlencoded",
            "Dnt": "1",
            "Host": "savingsbonds.gov",
            "Origin": "https://savingsbonds.gov",
            "Referer": "https://savingsbonds.gov/GA-FI/FedInvest/selectSecurityPriceDate",
            "Sec-Ch-Ua": '"Google Chrome";v="125", "Chromium";v="125", "Not.A/Brand";v="24"',
            "Sec-Ch-Ua-Mobile": "?0",
            "Sec-Ch-Ua-Platform": '"Windows"',
            "Sec-Fetch-Dest": "document",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-User": "?1",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.36",
        }
        self._logger.debug(f"UST Prices - {date} Payload: {payload}")
        cols_to_return = ["cusip", "offer_price", "bid_price", "eod_price"]
        retries = 0
        try:
            while retries < max_retries:
                try:
                    url = "https://savingsbonds.gov/GA-FI/FedInvest/selectSecurityPriceDate"
                    response = await client.post(
                        url,
                        data=payload,
                        headers=headers,
                        follow_redirects=False,
                        timeout=self._global_timeout,
                    )
                    if response.is_redirect:
                        redirect_url = response.headers.get("Location")
                        self._logger.debug(f"UST Prices - {date} Redirecting to {redirect_url}")
                        response = await client.get(redirect_url, headers=headers)

                    response.raise_for_status()
                    tables = pd.read_html(response.content, header=0)
                    df = tables[0]
                    if cusips:
                        missing_cusips = [cusip for cusip in cusips if cusip not in df["CUSIP"].values]
                        if missing_cusips:
                            self._logger.warning(f"UST Prices Warning - The following CUSIPs are not found in the DataFrame: {missing_cusips}")
                    df = df[df["CUSIP"].isin(cusips)] if cusips else df
                    df.columns = df.columns.str.lower()
                    df = df.query("`security type` not in ['TIPS', 'MARKET BASED FRN']")
                    df = df.rename(
                        columns={
                            "buy": "offer_price",
                            "sell": "bid_price",
                            "end of day": "eod_price",
                        }
                    )
                    if uid:
                        return date, df[cols_to_return], uid
                    return date, df[cols_to_return]

                except httpx.HTTPStatusError as e:
                    self._logger.error(f"UST Prices - Bad Status for {date}: {response.status_code}")
                    if response.status_code == 404:
                        if uid:
                            return date, df[cols_to_return], uid
                        return date, df[cols_to_return]
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST Prices - Throttled. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

                except Exception as e:
                    self._logger.error(f"UST Prices - Error for {date}: {e}")
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST Prices - Throttled. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

            raise ValueError(f"UST Prices - Max retries exceeded for {date}")
        except Exception as e:
            self._logger.error(e)
            if uid:
                return date, pd.DataFrame(columns=cols_to_return), uid
            return date, pd.DataFrame(columns=cols_to_return)

    async def _fetch_cusip_prices_fedinvest_with_semaphore(self, semaphore, *args, **kwargs):
        async with semaphore:
            return await self._fetch_cusip_prices_fedinvest(*args, **kwargs)

    async def _build_fetch_tasks_cusip_prices_fedinvest(
        self,
        client: httpx.AsyncClient,
        dates: List[datetime],
        cusips: Optional[List[str]] = None,
        uid: Optional[str | int] = None,
        max_concurrent_tasks: int = 64,
    ):
        semaphore = asyncio.Semaphore(max_concurrent_tasks)
        tasks = [
            self._fetch_cusip_prices_fedinvest_with_semaphore(
                semaphore,
                client=client,
                date=date,
                cusips=cusips,
                uid=uid,
            )
            for date in dates
        ]
        return tasks

    # we store FedInvest prices in github b/c it speeds up fetching by a ton!
    def _github_headers(self, path: str):
        return {
            "authority": "raw.githubusercontent.com",
            "method": "GET",
            "path": path,
            "scheme": "https",
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9",
            "cache-control": "no-cache",
            "dnt": "1",
            "pragma": "no-cache",
            "priority": "u=0, i",
            "sec-ch-ua": '"Not)A;Brand";v="99", "Google Chrome";v="127", "Chromium";v="127"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "none",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36",
        }

    async def _fetch_cusip_prices_from_github(
        self,
        client: httpx.AsyncClient,
        date: datetime,
        cusips: Optional[List[str]] = None,
        uid: Optional[str | int] = None,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
        cols_to_return: Optional[List[str]] = [
            "cusip",
            "offer_yield",
            "bid_yield",
            "eod_yield",
        ],
        cusip_ref_replacement_dict: Optional[Dict[str, str]] = None,
        return_transpose_df: Optional[bool] = False,
        assume_otrs: Optional[bool] = False,
        set_cusips_as_index: Optional[bool] = False,
    ):
        date_str = date.strftime("%Y-%m-%d")
        headers = self._github_headers(path=f"/cleeclee123/CUSIP-Set/main/{date_str}.json")
        url = f"https://raw.githubusercontent.com/cleeclee123/CUSIP-Set/main/{date_str}.json"
        retries = 0
        cols_to_return_copy = cols_to_return.copy()
        try:
            while retries < max_retries:
                try:
                    response = await client.get(url, headers=headers)
                    response.raise_for_status()
                    res_json = response.json()
                    df = pd.DataFrame(res_json["data"])
                    if df.empty:
                        self._logger.error(f"UST Prices GitHub - Data is Empty for {date}")
                        if uid:
                            return date, None, uid
                        return date, None

                    df["issue_date"] = pd.to_datetime(df["issue_date"])
                    df["maturity_date"] = pd.to_datetime(df["maturity_date"])

                    if cusips:
                        missing_cusips = [cusip for cusip in cusips if cusip not in df["cusip"].values]
                        if missing_cusips:
                            self._logger.warning(f"UST Prices Warning - The following CUSIPs are not found in the DataFrame: {missing_cusips}")
                    df = df[df["cusip"].isin(cusips)] if cusips else df

                    if cusip_ref_replacement_dict:
                        df["cusip"] = df["cusip"].replace(cusip_ref_replacement_dict)

                    if assume_otrs and "original_security_term" in df.columns:
                        df = df.sort_values(by=["issue_date"], ascending=False)
                        df = df.groupby("original_security_term").first().reset_index()
                        cusip_to_term_dict = dict(zip(df["cusip"], df["original_security_term"]))
                        df["cusip"] = df["cusip"].replace(cusip_to_term_dict)

                    if set_cusips_as_index:
                        df = df.set_index("cusip")
                        cols_to_return_copy.remove("cusip")

                    if return_transpose_df:
                        df = df[cols_to_return_copy].T
                    else:
                        df = df[cols_to_return_copy]

                    if uid:
                        return date, df, uid
                    return date, df

                except httpx.HTTPStatusError as e:
                    self._logger.error(f"UST Prices GitHub - Error for {date}: {e}")
                    if response.status_code == 404:
                        if uid:
                            return date, None, uid
                        return date, None
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST Prices GitHub - Throttled for {date}. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)
                except Exception as e:
                    self._logger.error(f"UST Prices GitHub - Error for {date}: {e}")
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST Prices GitHub - Throttled for {date}. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

            raise ValueError(f"UST Prices GitHub - Max retries exceeded for {date}")

        except Exception as e:
            self._logger.error(e)
            if uid:
                return date, None, uid
            return date, None

    async def _fetch_cusip_prices_from_github_with_semaphore(self, semaphore, *args, **kwargs):
        async with semaphore:
            return await self._fetch_cusip_prices_from_github(*args, **kwargs)

    async def _build_fetch_tasks_cusip_prices_from_github(
        self,
        client: httpx.AsyncClient,
        dates: List[datetime],
        cusips: Optional[List[str]] = None,
        uid: Optional[str | int] = None,
        max_concurrent_tasks: int = 64,
        my_semaphore: Optional[asyncio.Semaphore] = None,
        cols_to_return: Optional[List[str]] = [
            "cusip",
            "offer_yield",
            "bid_yield",
            "eod_yield",
        ],
    ):
        semaphore = my_semaphore or asyncio.Semaphore(max_concurrent_tasks)
        tasks = [
            self._fetch_cusip_prices_from_github_with_semaphore(
                semaphore,
                client=client,
                date=date,
                cusips=cusips,
                uid=uid,
                cols_to_return=cols_to_return,
            )
            for date in dates
        ]
        return tasks

    # ct yields are yields of OTR USTs at respective tenors timeseres
    def get_historical_ct_yields(
        self,
        start_date: datetime,
        end_date: datetime,
        tenors: Optional[List[str]] = None,
        use_bid_side: Optional[bool] = False,
        use_offer_side: Optional[bool] = False,
        use_mid_side: Optional[bool] = False,
    ):
        side = "eod"
        if use_bid_side:
            side = "bid"
        elif use_offer_side:
            side = "offer"
        elif use_mid_side:
            side = "mid"

        url = f"https://raw.githubusercontent.com/cleeclee123/CUSIP-Timeseries/main/historical_ct_yields_{side}_side.json"
        try:
            res = requests.get(
                url,
                headers=self._github_headers(path=f"/cleeclee123/CUSIP-Timeseries/main/historical_ct_yields_{side}_side.json"),
                proxies=self._proxies,
            )
            res.raise_for_status()
            df = pd.DataFrame(res.json())
            df["Date"] = pd.to_datetime(df["Date"])
            df = df[(df["Date"] >= start_date) & (df["Date"] <= end_date)]
            df = df.reset_index(drop=True)
            df.columns = ["Date", "CT1M", "CT2M", "CT3M", "CT4M", "CT6M", "CT1", "CT2", "CT3", "CT5", "CT7", "CT10", "CT20", "CT30"]

            if tenors:
                tenors = ["Date"] + tenors
                return df[tenors]
            return df
        except Exception as e:
            self._logger.error(f"Historical CT Yields GitHub - {str(e)}")

    async def _fetch_historical_ct_yields(
        self,
        client: httpx.AsyncClient,
        start_date: datetime,
        end_date: datetime,
        tenors: Optional[List[str]] = None,
        use_bid_side: Optional[bool] = False,
        use_offer_side: Optional[bool] = False,
        use_mid_side: Optional[bool] = False,
    ):
        side = "eod"
        if use_bid_side:
            side = "bid"
        elif use_offer_side:
            side = "offer"
        elif use_mid_side:
            side = "mid"

        url = f"https://raw.githubusercontent.com/cleeclee123/CUSIP-Timeseries/main/historical_ct_yields_{side}_side.json"

        try:
            res = await client.get(
                url,
                headers=self._github_headers(path=f"/cleeclee123/CUSIP-Timeseries/main/historical_ct_yields_{side}_side.json"),
            )
            res.raise_for_status()
            df = pd.DataFrame(res.json())
            df["Date"] = pd.to_datetime(df["Date"])
            df = df[(df["Date"] >= start_date) & (df["Date"] <= end_date)]
            df = df.reset_index(drop=True)
            df.columns = ["Date", "CT2M", "CT3M", "CT6M", "CT1", "CT2", "CT3", "CT5", "CT7", "CT10", "CT20", "CT30"]

            if tenors:
                tenors = ["Date"] + tenors
                return df[tenors]
            return df
        except httpx.HTTPStatusError as e:
            self._logger.error(f"HTTP Error: {str(e)}")
        except Exception as e:
            self._logger.error(f"Historical CT Yields GitHub - {str(e)}")

    # We store cusip timeseries data on github
    async def _fetch_cusip_timeseries_github(
        self,
        client: httpx.AsyncClient,
        cusip: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        uid: Optional[str | int] = None,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
    ):
        retries = 0
        try:
            while retries < max_retries:
                url = f"https://raw.githubusercontent.com/cleeclee123/CUSIP-Timeseries/main/{cusip}.json"
                headers = self._github_headers(path=f"/cleeclee123/CUSIP-Timeseries/main/{cusip}.json")
                try:
                    response = await client.get(url, headers=headers)
                    response.raise_for_status()
                    response_json = response.json()
                    df = pd.DataFrame(response_json)
                    df["Date"] = pd.to_datetime(df["Date"])
                    df = df.sort_values(by="Date")
                    if start_date:
                        df = df[df["Date"].dt.date >= start_date.date()]
                    if end_date:
                        df = df[df["Date"].dt.date <= end_date.date()]
                    if uid:
                        return cusip, df, uid
                    return cusip, df

                except httpx.HTTPStatusError as e:
                    self._logger.error(f"UST Timeseries GitHub - Error for {cusip}: {e}")
                    if response.status_code == 404:
                        if uid:
                            return cusip, None, uid
                        return cusip, None
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST Timeseries GitHub - Throttled for {cusip}. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)
                except Exception as e:
                    self._logger.error(f"UST Timeseries GitHub - Error for {cusip}: {e}")
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST Timeseries GitHub - Throttled for {cusip}. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

            raise ValueError(f"UST Timeseries GitHub - Max retries exceeded for {cusip}")

        except Exception as e:
            self._logger.error(e)
            if uid:
                return cusip, None, uid
            return cusip, None

    async def _fetch_cusip_timeseries_github_with_semaphore(self, semaphore, *args, **kwargs):
        async with semaphore:
            return await self._fetch_cusip_timeseries_github(*args, **kwargs)

    def cusips_timeseries(
        self,
        cusips: List[str],
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        max_concurrent_tasks: int = 64,
        max_keepalive_connections: int = 5,
        to_df_cols: Optional[List[str]] = None,
        use_dict_key_df_cols: Optional[bool] = False,
        df_col_delimitter: Optional[str] = "-",
    ):
        async def build_tasks(
            client: httpx.AsyncClient,
            cusips: List[str],
        ):
            tasks = []
            semaphore = asyncio.Semaphore(max_concurrent_tasks)
            for cusip in cusips:
                task = self._fetch_cusip_timeseries_github_with_semaphore(
                    semaphore=semaphore,
                    client=client,
                    start_date=start_date,
                    end_date=end_date,
                    cusip=cusip,
                )
                tasks.append(task)

            return await asyncio.gather(*tasks)

        async def run_fetch_all(cusips: List[str]):
            limits = httpx.Limits(
                max_connections=max_concurrent_tasks,
                max_keepalive_connections=max_keepalive_connections,
            )
            async with httpx.AsyncClient(limits=limits) as client:
                all_data = await build_tasks(client=client, cusips=cusips)
                return all_data

        results: List[Tuple[str, pd.DataFrame]] = asyncio.run(run_fetch_all(cusips=cusips))
        results_dict: Dict[str, pd.DataFrame] = {dt: df for dt, df in results if dt is not None and df is not None}

        if to_df_cols:
            renamed_dfs = []
            for key, df in results_dict.items():
                temp_df = df[["Date"] + to_df_cols].copy()
                if use_dict_key_df_cols:
                    temp_df = temp_df.rename(columns={col: f"{key}" for col in temp_df.columns if col != "Date"})
                else:
                    temp_df = temp_df.rename(columns={col: f"{key}{df_col_delimitter}{col}" for col in temp_df.columns if col != "Date"})
                renamed_dfs.append(temp_df)

            from functools import reduce

            merged_df = reduce(lambda left, right: pd.merge(left, right, on="Date", how="outer"), renamed_dfs)
            return merged_df

        return results_dict



import numpy as np


class BjorkChristensenAugmentedCurve:
    def __init__(
        self,
        beta0: float,
        beta1: float,
        beta2: float,
        beta3: float,
        beta4: float,
        tau: float,
    ) -> None:
        self.beta0 = beta0
        self.beta1 = beta1
        self.beta2 = beta2
        self.beta3 = beta3
        self.beta4 = beta4
        self.tau = tau

    def _time_decay(self, t: np.ndarray) -> np.ndarray:
        return self.beta1 * (t / (2 * self.tau))

    def _hump(self, t: np.ndarray) -> np.ndarray:
        return self.beta2 * ((1 - np.exp(-t / self.tau)) / (t / self.tau))

    def _second_hump(self, t: np.ndarray) -> np.ndarray:
        return self.beta3 * (
            (1 - np.exp(-t / self.tau)) / (t / self.tau) - np.exp(-t / self.tau)
        )

    def _third_hump(self, t: np.ndarray) -> np.ndarray:
        return self.beta4 * ((1 - np.exp(-2 * t / self.tau)) / (2 * t / self.tau))

    def factor_matrix(self, t: np.ndarray) -> np.ndarray:
        """Create the factor matrix for OLS."""
        factor1 = self._time_decay(t)
        factor2 = self._hump(t)
        factor3 = self._second_hump(t)
        factor4 = self._third_hump(t)
        constant = np.ones(t.shape)
        return np.column_stack([constant, factor1, factor2, factor3, factor4])

    def zero(self, t: np.ndarray) -> np.ndarray:
        """Calculate the zero rate for a given maturity."""
        return (
            self.beta0
            + self._time_decay(t)
            + self._hump(t)
            + self._second_hump(t)
            + self._third_hump(t)
        )

    def __call__(self, t: np.ndarray) -> np.ndarray:
        """Calculate the zero rate when the instance is called directly."""
        return self.zero(t)



from typing import Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from ..utils.simulation import Simulation


class Vasicek:

    def __init__(
        self,
        alpha: float,
        beta: float,
        sigma: float,
        rt: float,
        time: float,
        delta_time: float,
    ) -> None:
        self._alpha = alpha
        self._beta = beta
        self._sigma = sigma
        self._rt = rt
        self._dt = delta_time
        self._steps = int(time / delta_time)

    def get_attr(self, attr: str) -> Union[float, int]:
        return self.__getattribute__(attr)

    def _sigma_part(self, n: int) -> float:
        return (
            self.get_attr("_sigma")
            * np.sqrt(self.get_attr("_dt"))
            * np.random.normal(size=n)
        )

    def _mu_dt(self, rt: np.ndarray) -> float:
        return (
            self.get_attr("_alpha")
            * (self.get_attr("_beta") - rt)
            * self.get_attr("_dt")
        )

    def simulate_paths(self, n: int) -> np.array:
        simulation = np.zeros(shape=(self.get_attr("_steps"), n))
        simulation[0, :] = self._rt
        for i in range(1, self.get_attr("_steps"), 1):
            dr = self._mu_dt(simulation[i - 1, :]) + self._sigma_part(n)
            simulation[i, :] = simulation[i - 1, :] + dr
        return Simulation(simulation, self.get_attr("_dt"))

    @staticmethod
    def plot_calibrated(
        simul: Simulation,
        instantaneous_forward: Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]],
    ) -> None:
        fig = plt.figure(figsize=(12.5, 8))
        fig.suptitle("Model Fitting Curve T=0")
        fig.canvas.set_window_title("Model Fitting Curve T=0")
        ax1 = fig.add_subplot(111)
        ax1.set_xlabel("t, years")
        ax1.set_ylabel("Yield")
        ax1.plot(
            np.linspace(1, simul.get_steps, simul.get_steps) * simul.get_dt,
            simul.get_sim,
            lw=0.5,
        )
        ax1.plot(
            np.linspace(1, simul.get_steps, simul.get_steps) * simul.get_dt,
            simul.yield_curve().get_rate,
            lw=3,
            c="Navy",
            label="Vasicek Term Structure",
        )
        ax1.plot(
            instantaneous_forward.get_time,
            instantaneous_forward.get_rate,
            c="darkred",
            label="Initial Term Structure",
            lw=3,
        )
        plt.legend()
        plt.show()
        return fig



import asyncio
import math
import os
import time
import warnings
from datetime import datetime
from typing import Dict, List, Optional

import httpx
import numpy as np
import pandas as pd
import requests

from CurvyCUSIPs.DataFetcher.base import DataFetcherBase
from CurvyCUSIPs.utils.ust_utils import JSON, build_treasurydirect_header, get_active_cusips, historical_auction_cols, last_day_n_months_ago

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)

import asyncio
import http
import os
import shutil
from typing import Dict, List, Optional, TypeAlias

import aiohttp
import pandas as pd

JSON: TypeAlias = dict[str, "JSON"] | list["JSON"] | str | int | float | bool | None


def get_treasurygov_header(year: int, cj: http.cookies = None) -> Dict[str, str]:
    cookie_str = ""
    if cj:
        cookies = {cookie.name: cookie.value for cookie in cj if "home.treasury.gov" in cookie.domain}
        cookie_str = "; ".join([f"{key}={value}" for key, value in cookies.items()])

    headers = {
        "authority": "home.treasury.gov",
        "method": "GET",
        "path": f"/resource-center/data-chart-center/interest-rates/TextView?type=daily_treasury_yield_curve&field_tdr_date_value={year}",
        "scheme": "https",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
        "Accept-Encoding": "gzip, deflate, br",
        "Accept-Language": "en-US,en;q=0.9",
        "Cache-Control": "max-age=0",
        "Cookie": cookie_str,
        "Dnt": "1",
        "Sec-Ch-Ua": '"Chromium";v="116", "Not)A;Brand";v="24", "Google Chrome";v="116"',
        "Sec-Ch-Ua-Mobile": "?0",
        "Sec-Ch-Ua-Platform": "Windows",
        "Sec-Fetch-Dest": "document",
        "Sec-Fetch-Mode": "navigate",
        "Sec-Fetch-Site": "none",
        "Sec-Fetch-User": "?1",
        "Upgrade-Insecure-Requests": "1",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
    }

    if cookie_str == "":
        del headers["Cookie"]

    return headers


def multi_download_year_treasury_par_yield_curve_rate(
    years: List[int],
    raw_path: str,
    download=False,
    real_par_yields=False,
    cj: http.cookies = None,
    run_all=False,
    verbose=False,
    proxy: Optional[str] = None,
) -> pd.DataFrame:
    async def fetch_from_treasurygov(session: aiohttp.ClientSession, url: str, curr_year: int) -> pd.DataFrame:
        try:
            headers = get_treasurygov_header(curr_year, cj)
            treasurygov_data_type = "".join(url.split("?type=")[1].split("&field")[0])
            full_file_path = os.path.join(raw_path, "temp", f"{treasurygov_data_type}.csv")
            async with session.get(url, headers=headers, proxy=proxy) as response:
                if response.status == 200:
                    with open(full_file_path, "wb") as f:
                        chunk_size = 8192
                        while True:
                            chunk = await response.content.read(chunk_size)
                            if not chunk:
                                break
                            f.write(chunk)
                    return {treasurygov_data_type: await convert_csv_to_excel(full_file_path)}
                else:
                    raise Exception(f"Bad Status: {response.status}")
        except Exception as e:
            print(e) if verbose else None
            return {treasurygov_data_type: pd.DataFrame()}

    async def convert_csv_to_excel(full_file_path: str | None) -> str:
        if not full_file_path:
            return

        copy = full_file_path
        rdir_path = copy.split("\\")
        rdir_path.remove("temp")
        renamed = str.join("\\", rdir_path)
        renamed = f"{renamed.split('.')[0]}.xlsx"

        df_temp = pd.read_csv(full_file_path)
        df_temp["Date"] = pd.to_datetime(df_temp["Date"])
        df_temp["Date"] = df_temp["Date"].dt.strftime("%Y-%m-%d")
        if download:
            df_temp.to_excel(f"{renamed.split('.')[0]}.xlsx", index=False)
        os.remove(full_file_path)
        return df_temp

    async def get_promises(session: aiohttp.ClientSession):
        tasks = []
        for year in years:
            daily_par_yield_curve_url = f"https://home.treasury.gov/resource-center/data-chart-center/interest-rates/daily-treasury-rates.csv/{year}/all?type=daily_treasury_yield_curve&field_tdr_date_value={year}&page&_format=csv"
            daily_par_real_yield_curve_url = f"https://home.treasury.gov/resource-center/data-chart-center/interest-rates/daily-treasury-rates.csv/{year}/all?type=daily_treasury_real_yield_curve&field_tdr_date_value={year}&amp;page&amp;_format=csv"
            daily_treasury_bill_rates_url = f"https://home.treasury.gov/resource-center/data-chart-center/interest-rates/daily-treasury-rates.csv/{year}/all?type=daily_treasury_bill_rates&field_tdr_date_value={year}&page&_format=csv"
            daily_treaury_long_term_rates_extrapolation_factors_url = f"https://home.treasury.gov/resource-center/data-chart-center/interest-rates/daily-treasury-rates.csv/{year}/all?type=daily_treasury_long_term_rate&field_tdr_date_value={year}&page&_format=csv"
            daily_treasury_real_long_term_rates_averages = f"https://home.treasury.gov/resource-center/data-chart-center/interest-rates/daily-treasury-rates.csv/{year}/all?type=daily_treasury_real_long_term&field_tdr_date_value={year}&page&_format=csv"
            if run_all:
                tasks.extend(
                    [
                        fetch_from_treasurygov(session, daily_par_yield_curve_url, year),
                        fetch_from_treasurygov(session, daily_par_real_yield_curve_url, year),
                        fetch_from_treasurygov(session, daily_treasury_bill_rates_url, year),
                        fetch_from_treasurygov(
                            session,
                            daily_treaury_long_term_rates_extrapolation_factors_url,
                            year,
                        ),
                        fetch_from_treasurygov(session, daily_treasury_real_long_term_rates_averages, year),
                    ]
                )
            else:
                curr_url = daily_par_yield_curve_url if not real_par_yields else daily_par_real_yield_curve_url
                task = fetch_from_treasurygov(session, curr_url, year)
                tasks.append(task)

        return await asyncio.gather(*tasks)

    async def run_fetch_all() -> List[pd.DataFrame]:
        async with aiohttp.ClientSession() as session:
            all_data = await get_promises(session)
            return all_data

    try:
        os.mkdir(os.path.join(raw_path, "temp"))
        dfs: List[Dict[str, pd.DataFrame]] = asyncio.run(run_fetch_all())
        shutil.rmtree(os.path.join(raw_path, "temp"))

        if not run_all:
            dfs = [next(iter(dictionary.values())) for dictionary in dfs]
            yield_df = pd.concat(dfs, ignore_index=True)
            return yield_df

        organized_by_ust_type_dict: Dict[str, List[pd.DataFrame]] = {}
        for dictionary in dfs:
            ust_data_type, df = next(iter(dictionary)), next(iter(dictionary.values()))
            if not ust_data_type or df is None or df.empty:
                continue
            if ust_data_type not in organized_by_ust_type_dict:
                organized_by_ust_type_dict[ust_data_type] = []
            organized_by_ust_type_dict[ust_data_type].append(df)

        organized_by_ust_type_df_dict_concated: Dict[str, pd.DataFrame] = {}
        for ust_data_type in organized_by_ust_type_dict.keys():
            dfs = organized_by_ust_type_dict[ust_data_type]
            concated_df = pd.concat(dfs, ignore_index=True)
            organized_by_ust_type_df_dict_concated[ust_data_type] = concated_df

        return organized_by_ust_type_df_dict_concated
    except Exception as e:
        if os.path.exists(os.path.join(raw_path, "temp")):
            shutil.rmtree(os.path.join(raw_path, "temp"))
        raise e


class USTreasuryDataFetcher(DataFetcherBase):
    _use_ust_issue_date: bool = False
    _historical_auctions_df: pd.DataFrame = (None,)
    _valid_tenors_strings = [
        "CMT1M",
        "CMT2M",
        "CMT3M",
        "CMT4M",
        "CMT6M",
        "CMT1",
        "CMT2",
        "CMT3",
        "CMT5",
        "CMT7",
        "CMT10",
        "CMT20",
        "CMT30",
    ]

    def __init__(
        self,
        use_ust_issue_date: Optional[bool] = False,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        super().__init__(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self._use_ust_issue_date = use_ust_issue_date
        time.sleep(0.01)
        self._historical_auctions_df = self.get_auctions_df()
        self._historical_auctions_df["auction_date"] = pd.to_datetime(self._historical_auctions_df["auction_date"])
        self._historical_auctions_df["issue_date"] = pd.to_datetime(self._historical_auctions_df["issue_date"])
        self._historical_auctions_df["maturity_date"] = pd.to_datetime(self._historical_auctions_df["maturity_date"])
        self._historical_auctions_df["int_rate"] = pd.to_numeric(self._historical_auctions_df["int_rate"], errors="coerce")
        self._historical_auctions_df["high_investment_rate"] = pd.to_numeric(self._historical_auctions_df["high_investment_rate"], errors="coerce")
        self._historical_auctions_df["ust_label"] = self._historical_auctions_df.apply(
            lambda row: (
                f"{row['high_investment_rate']:.3f}% {row['maturity_date'].strftime('%b-%y')}"
                if pd.isna(row["int_rate"])
                else f"{row['int_rate']:.3f}% {row['maturity_date'].strftime('%b-%y')}"
            ),
            axis=1,
        )

    # ust label: f"{coupon}s {datetime.strftime("%Y-%m-%d")}"
    def cme_ust_label_to_cusip(self, ust_label: str):
        try:
            coupon = float(ust_label.split("s")[0])
            maturity = ust_label.split(" ")[1]
            ust_row = self._historical_auctions_df[
                (self._historical_auctions_df["int_rate"] == coupon) & (self._historical_auctions_df["maturity_date"] == maturity)
            ]
            return ust_row.to_dict("records")[0]
        except:
            raise Exception("LABEL NOT FOUND")

    def cusip_to_cme_ust_label(self, cusip: str):
        ust_row = self._historical_auctions_df[self._historical_auctions_df["cusip"] == cusip].to_dict("records")[0]
        return f"{ust_row["int_rate"]}s {ust_row["maturity_date"]}"

    # ust_label = f"{row['int_rate']:.3f}% {row['maturity_date'].strftime('%b-%y')}"
    def ust_label_to_cusip(self, ust_label: str, return_all_occurrences=False):
        try:
            ust_row = self._historical_auctions_df[self._historical_auctions_df["ust_label"] == ust_label]
            occurences = ust_row.to_dict("records")
            if return_all_occurrences:
                return occurences
            return occurences[-1]  # return the earliest issue date by default
        except Exception as e:
            if ust_row.empty:
                raise Exception(f"LABEL NOT FOUND - {e}")
            else:
                raise Exception(f"SOMETHING WENT WRONG - {e}")

    def cusip_to_ust_label(self, cusip: str, return_all_occurrences=False):
        try:
            ust_row = self._historical_auctions_df[self._historical_auctions_df["cusip"] == cusip]
            occurences = ust_row.to_dict("records")
            if return_all_occurrences:
                return occurences
            return occurences[-1]  # return the earliest issue date by default
        except Exception as e:
            if ust_row.empty:
                raise Exception(f"CUSIP NOT FOUND - {e}")
            else:
                raise Exception(f"SOMETHING WENT WRONG - {e}")

    async def _build_fetch_tasks_historical_treasury_auctions(
        self,
        client: httpx.AsyncClient,
        assume_data_size=True,
        uid: Optional[str | int] = None,
        return_df: Optional[bool] = False,
        as_of_date: Optional[datetime] = None,  # active cusips as of
        max_retries: Optional[int] = 5,
        backoff_factor: Optional[int] = 1,
    ):
        MAX_TREASURY_GOV_API_CONTENT_SIZE = 10000
        NUM_REQS_NEEDED_TREASURY_GOV_API = 2

        def get_treasury_query_sizing() -> List[str]:
            base_url = "https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/od/auctions_query?page[number]=1&page[size]=1"
            res = requests.get(base_url, headers=build_treasurydirect_header(), proxies=self._proxies)
            if res.ok:
                meta = res.json()["meta"]
                size = meta["total-count"]
                number_requests = math.ceil(size / MAX_TREASURY_GOV_API_CONTENT_SIZE)
                return [
                    f"https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/od/auctions_query?page[number]={i+1}&page[size]={MAX_TREASURY_GOV_API_CONTENT_SIZE}"
                    for i in range(0, number_requests)
                ]
            else:
                raise ValueError(f"UST Auctions - Query Sizing Bad Status: ", {res.status_code})

        links = (
            get_treasury_query_sizing()
            if not assume_data_size
            else [
                f"https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/od/auctions_query?page[number]={i+1}&page[size]={MAX_TREASURY_GOV_API_CONTENT_SIZE}"
                for i in range(0, NUM_REQS_NEEDED_TREASURY_GOV_API)
            ]
        )
        self._logger.debug(f"UST Auctions - Number of Links to Fetch: {len(links)}")
        self._logger.debug(f"UST Auctions - Links: {links}")

        async def fetch(
            client: httpx.AsyncClient,
            url,
            as_of_date: Optional[datetime] = None,
            return_df: Optional[bool] = False,
            uid: Optional[str | int] = None,
            max_retries: Optional[int] = 5,
            backoff_factor: Optional[int] = 1,
        ):
            retries = 0
            try:
                while retries < max_retries:
                    try:
                        response = await client.get(
                            url,
                            headers=build_treasurydirect_header(),
                        )
                        response.raise_for_status()
                        json_data: JSON = response.json()
                        if as_of_date:
                            df = get_active_cusips(
                                auction_json=json_data["data"],
                                as_of_date=as_of_date,
                                use_issue_date=self._use_ust_issue_date,
                            )
                            if uid:
                                return df[historical_auction_cols()], uid
                            return df[historical_auction_cols()]

                        if return_df and not as_of_date:
                            if uid:
                                return (
                                    pd.DataFrame(json_data["data"])[historical_auction_cols()],
                                    uid,
                                )
                            return pd.DataFrame(json_data["data"])[historical_auction_cols()]
                        if uid:
                            return json_data["data"], uid
                        return json_data["data"]

                    except httpx.HTTPStatusError as e:
                        self._logger.error(f"UST AUCTIONS - Bad Status: {response.status_code}")

                        # UST's website is very weird sometimes - maybe getting throttled?
                        # if response.status_code == 404:
                        #     if uid:
                        #         return pd.DataFrame(columns=historical_auction_cols()), uid
                        #     return pd.DataFrame(columns=historical_auction_cols())

                        retries += 1
                        wait_time = backoff_factor * (2 ** (retries - 1))
                        self._logger.debug(f"UST AUCTIONS - Throttled. Waiting for {wait_time} seconds before retrying...")
                        await asyncio.sleep(wait_time)

                    except Exception as e:
                        self._logger.error(f"UST AUCTIONS - Error: {e}")
                        retries += 1
                        wait_time = backoff_factor * (2 ** (retries - 1))
                        self._logger.debug(f"UST STRIPPING Activity - Throttled. Waiting for {wait_time} seconds before retrying...")
                        await asyncio.sleep(wait_time)

                raise ValueError(f"UST AUCTIONS Activity - Max retries exceeded")

            except Exception as e:
                self._logger.error(e)
                raise e
                # if uid:
                #     return pd.DataFrame(columns=historical_auction_cols()), uid
                # return pd.DataFrame(columns=historical_auction_cols())

        tasks = [
            fetch(client=client, url=url, as_of_date=as_of_date, return_df=return_df, uid=uid, max_retries=max_retries, backoff_factor=backoff_factor)
            for url in links
        ]
        return tasks

    def get_auctions_df(self, as_of_date: Optional[datetime] = None) -> pd.DataFrame:
        async def build_tasks(client: httpx.AsyncClient, as_of_date: datetime):
            tasks = await self._build_fetch_tasks_historical_treasury_auctions(client=client, as_of_date=as_of_date, return_df=True)
            return await asyncio.gather(*tasks)

        async def run_fetch_all(as_of_date: datetime):
            async with httpx.AsyncClient(proxy=self._proxies["https"]) as client:
                all_data = await build_tasks(client=client, as_of_date=as_of_date)
                return all_data

        dfs = asyncio.run(run_fetch_all(as_of_date=as_of_date))
        auctions_df: pd.DataFrame = pd.concat(dfs)
        auctions_df = auctions_df.sort_values(by=["auction_date"], ascending=False)
        return auctions_df

    def get_historical_cmt_yields(
        self,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        treasury_data_dir: Optional[str] = None,
        download_treasury_par_yields: Optional[bool] = False,
        apply_long_term_extrapolation_factor: Optional[bool] = False,
        tenors: Optional[List[str]] = None,
    ):
        print("Fetching from treasury.gov...")
        dir_path = treasury_data_dir or os.getcwd()
        start_year = start_date.year if start_date else 1990
        end_year = end_date.year if end_date else 2024
        if start_year == end_year:
            start_year = start_year - 1
        years = [str(x) for x in range(end_year, start_year, -1)]
        ust_daily_data = multi_download_year_treasury_par_yield_curve_rate(
            years,
            dir_path,
            run_all=True,
            download=download_treasury_par_yields,
            proxy=self._proxies["https"],
        )
        if "daily_treasury_yield_curve" not in ust_daily_data:
            raise ValueError("CMT Yield - Fetch Failed - Fetching from treasury.gov")
        df_par_rates: pd.DataFrame = ust_daily_data["daily_treasury_yield_curve"]
        df_par_rates["Date"] = pd.to_datetime(df_par_rates["Date"])
        df_par_rates = df_par_rates.sort_values(by=["Date"], ascending=True).reset_index(drop=True)

        if start_date:
            df_par_rates = df_par_rates[df_par_rates["Date"] >= start_date]
        if end_date:
            df_par_rates = df_par_rates[df_par_rates["Date"] <= end_date]

        if apply_long_term_extrapolation_factor:
            try:
                df_lt_avg_rate = ust_daily_data["daily_treasury_long_term_rate"]
                df_lt_avg_rate["Date"] = pd.to_datetime(df_lt_avg_rate["Date"])
                df_par_rates_lt_adj = pd.merge(df_par_rates, df_lt_avg_rate, on="Date", how="left")
                df_par_rates_lt_adj["20 Yr"] = df_par_rates_lt_adj["20 Yr"].fillna(df_par_rates_lt_adj["TREASURY 20-Yr CMT"])
                df_par_rates_lt_adj["30 Yr"] = np.where(
                    df_par_rates_lt_adj["30 Yr"].isna(),
                    df_par_rates_lt_adj["20 Yr"] + df_par_rates_lt_adj["Extrapolation Factor"],
                    df_par_rates_lt_adj["30 Yr"],
                )
                df_par_rates_lt_adj = df_par_rates_lt_adj.drop(
                    columns=[
                        "LT COMPOSITE (>10 Yrs)",
                        "TREASURY 20-Yr CMT",
                        "Extrapolation Factor",
                    ]
                )
                df_par_rates_lt_adj.columns = ["Date"] + self._valid_tenors_strings
                if tenors:
                    tenors = ["Date"] + tenors
                    df_par_rates_lt_adj[tenors]
                if download_treasury_par_yields:
                    df_par_rates_lt_adj.to_excel(
                        os.path.join(
                            dir_path,
                            "daily_treasury_par_yields_with_lt_extrap.xlsx",
                        )
                    )
                return df_par_rates_lt_adj
            except Exception as e:
                self._logger.error(f"UST CMT Yields - LT Extra Failed: {str(e)}")
                msg = "Applying Long-Term Extrapolation Factor Failed"
                is_to_recent = start_date > datetime(2006, 2, 9)
                if is_to_recent:
                    msg += f" - {start_date} is too recent - pick a starting date older than February 9, 2006"
                print(msg)

        df_par_rates.columns = ["Date"] + self._valid_tenors_strings

        if tenors:
            tenors = ["Date"] + tenors
            return df_par_rates[tenors]

        return df_par_rates

    async def _fetch_mspd_table_5(
        self,
        client: httpx.AsyncClient,
        date: datetime,
        uid: Optional[str | int] = None,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
    ):
        cols_to_return = [
            "cusip",
            "corpus_cusip",
            "outstanding_amt",  # not using this col since not all USTs have stripping activity - using MSPD table 3 for "outstanding_amt"
            "portion_unstripped_amt",
            "portion_stripped_amt",
            "reconstituted_amt",
        ]
        retries = 0
        try:
            while retries < max_retries:
                try:
                    last_n_months_last_business_days: List[datetime] = last_day_n_months_ago(date, n=2, return_all=True)
                    self._logger.debug(f"STRIPping - BDays: {last_n_months_last_business_days}")
                    dates_str_query = ",".join([date.strftime("%Y-%m-%d") for date in last_n_months_last_business_days])
                    url = f"https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/debt/mspd/mspd_table_5?filter=record_date:in:({dates_str_query})&page[number]=1&page[size]=10000"
                    self._logger.debug(f"STRIPping - {date} url: {url}")
                    response = await client.get(
                        url,
                        headers=build_treasurydirect_header(host_str="api.fiscaldata.treasury.gov"),
                    )
                    response.raise_for_status()
                    curr_stripping_activity_json = response.json()
                    curr_stripping_activity_df = pd.DataFrame(curr_stripping_activity_json["data"])
                    curr_stripping_activity_df = curr_stripping_activity_df[
                        curr_stripping_activity_df["security_class1_desc"] != "Treasury Inflation-Protected Securities"
                    ]
                    curr_stripping_activity_df["record_date"] = pd.to_datetime(curr_stripping_activity_df["record_date"], errors="coerce")
                    latest_date = curr_stripping_activity_df["record_date"].max()
                    curr_stripping_activity_df = curr_stripping_activity_df[curr_stripping_activity_df["record_date"] == latest_date]
                    curr_stripping_activity_df["outstanding_amt"] = pd.to_numeric(curr_stripping_activity_df["outstanding_amt"], errors="coerce")
                    curr_stripping_activity_df["portion_unstripped_amt"] = pd.to_numeric(
                        curr_stripping_activity_df["portion_unstripped_amt"],
                        errors="coerce",
                    )
                    curr_stripping_activity_df["portion_stripped_amt"] = pd.to_numeric(
                        curr_stripping_activity_df["portion_stripped_amt"], errors="coerce"
                    )
                    curr_stripping_activity_df["reconstituted_amt"] = pd.to_numeric(curr_stripping_activity_df["reconstituted_amt"], errors="coerce")
                    col1 = "cusip"
                    col2 = "security_class2_desc"
                    curr_stripping_activity_df.columns = [
                        col2 if col == col1 else col1 if col == col2 else col for col in curr_stripping_activity_df.columns
                    ]
                    curr_stripping_activity_df = curr_stripping_activity_df.rename(columns={"security_class2_desc": "corpus_cusip"})

                    if uid:
                        return date, curr_stripping_activity_df[cols_to_return], uid

                    return date, curr_stripping_activity_df[cols_to_return]

                except httpx.HTTPStatusError as e:
                    self._logger.error(f"STRIPping - Bad Status: {response.status_code}")
                    if response.status_code == 404:
                        if uid:
                            return date, pd.DataFrame(columns=cols_to_return), uid
                        return date, pd.DataFrame(columns=cols_to_return)

                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST STRIPPING Activity - Throttled. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

                except Exception as e:
                    self._logger.error(f"UST STRIPPING Activity - Error for {date}: {e}")
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"UST STRIPPING Activity - Throttled. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

            raise ValueError(f"UST STRIPPING Activity - Max retries exceeded for {date}")

        except Exception as e:
            self._logger.error(e)
            if uid:
                return date, pd.DataFrame(columns=cols_to_return), uid
            return date, pd.DataFrame(columns=cols_to_return)

    async def _fetch_mspd_table_5_with_semaphore(self, semaphore, *args, **kwargs):
        async with semaphore:
            return await self._fetch_mspd_table_5(*args, **kwargs)

    async def _build_fetch_tasks_historical_stripping_activity(
        self,
        client: httpx.AsyncClient,
        dates: List[datetime],
        uid: Optional[str | int] = None,
        minimize_api_calls=False,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
        max_concurrent_tasks: int = 64,
        my_semaphore: Optional[asyncio.Semaphore] = None,
    ):

        semaphore = my_semaphore or asyncio.Semaphore(max_concurrent_tasks)

        if minimize_api_calls:
            seen_month_years = set()
            filtered_list = []
            for date in dates:
                month_year = (date.year, date.month)
                if month_year not in seen_month_years:
                    seen_month_years.add(month_year)
                    filtered_list.append(date)

            tasks = [
                self._fetch_mspd_table_5_with_semaphore(
                    client=client, date=date, semaphore=semaphore, uid=uid, max_retries=max_retries, backoff_factor=backoff_factor
                )
                for date in filtered_list
            ]
            return tasks

        else:
            tasks = [
                self._fetch_mspd_table_5_with_semaphore(
                    client=client, date=date, semaphore=semaphore, uid=uid, max_retries=max_retries, backoff_factor=backoff_factor
                )
                for date in dates
            ]
            return tasks



from dataclasses import dataclass
from typing import Union, Tuple, Any
import numpy as np
from scipy.optimize import minimize
from numbers import Real

EPS = np.finfo(float).eps


@dataclass
class BjorkChristensenCurve:
    """Implementation of the Bjork-Christensen interest rate curve model."""

    beta0: float
    beta1: float
    beta2: float
    beta3: float
    tau: float

    def factors(
        self, T: Union[float, np.ndarray]
    ) -> Union[Tuple[float, float, float], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
        """Factor loadings for time(s) T, excluding constant."""
        T = np.maximum(T, EPS)  # Ensure T is not zero or negative to avoid instability

        factor1 = (1 - np.exp(-T / self.tau)) / (T / self.tau)
        factor2 = factor1 - np.exp(-T / self.tau)
        factor3 = (1 - np.exp(-2 * T / self.tau)) / (2 * T / self.tau)

        return factor1, factor2, factor3

    def factor_matrix(self, T: Union[float, np.ndarray]) -> np.ndarray:
        """Factor loadings for time(s) T as matrix columns, including constant column (=1.0)."""
        factor1, factor2, factor3 = self.factors(T)
        constant = np.ones_like(T)
        return np.stack([constant, factor1, factor2, factor3], axis=-1)

    def zero(self, T: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Zero rate(s) of this curve at time(s) T."""
        factor1, factor2, factor3 = self.factors(T)
        return (
            self.beta0
            + self.beta1 * factor1
            + self.beta2 * factor2
            + self.beta3 * factor3
        )

    def __call__(self, T: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Zero rate(s) of this curve at time(s) T."""
        return self.zero(T)

    def forward(self, T: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Instantaneous forward rate(s) of this curve at time(s) T."""
        exp_tt1 = np.exp(-T / self.tau)
        exp_tt2 = np.exp(-2 * T / self.tau)
        return (
            self.beta0
            + self.beta1 * exp_tt1
            + self.beta2 * exp_tt1 * T / self.tau
            + self.beta3 * exp_tt2 * T / (2 * self.tau)
        )



import asyncio
import warnings
from datetime import datetime
from typing import Dict, List, Optional, Tuple

import httpx
import pandas as pd
import requests

from CurvyCUSIPs.DataFetcher.base import DataFetcherBase

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)


class PublicDotcomDataFetcher(DataFetcherBase):
    _public_dotcom_jwt: str = None

    def __init__(
        self,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        super().__init__(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self._public_dotcom_jwt = self._fetch_public_dotcome_jwt()

    def _fetch_public_dotcome_jwt(self) -> str:
        try:
            jwt_headers = {
                "authority": "prod-api.154310543964.hellopublic.com",
                "method": "GET",
                "path": "/static/anonymoususer/credentials.json",
                "scheme": "https",
                "accept": "*/*",
                "accept-encoding": "gzip, deflate, br, zstd",
                "accept-language": "en-US,en;q=0.9",
                "cache-control": "no-cache",
                "content-type": "application/json",
                "dnt": "1",
                "origin": "https://public.com",
                "pragma": "no-cache",
                "priority": "u=1, i",
                "sec-ch-ua": '"Chromium";v="128", "Not;A=Brand";v="24", "Google Chrome";v="128"',
                "sec-ch-ua-mobile": "?0",
                "sec-ch-ua-platform": '"Windows"',
                "sec-fetch-dest": "empty",
                "sec-fetch-mode": "cors",
                "sec-fetch-site": "cross-site",
                "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36",
                "x-app-version": "web-1.0.9",
            }
            jwt_url = "https://prod-api.154310543964.hellopublic.com/static/anonymoususer/credentials.json"
            jwt_res = requests.get(jwt_url, headers=jwt_headers)
            jwt_str = jwt_res.json()["jwt"]
            return jwt_str
        except Exception as e:
            self._logger.error(f"Public.com JWT Request Failed: {e}")
            return None

    async def _fetch_cusip_timeseries_public_dotcom(
        self,
        client: httpx.AsyncClient,
        cusip: str,
        jwt_str: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
        uid: Optional[str | int] = None,
    ):
        cols_to_return = ["Date", "Price", "YTM"]  # YTW is same as YTM for cash USTs
        retries = 0
        try:
            if pd.isna(cusip) or not cusip:
                raise ValueError(f"Public.com - invalid CUSIP passed")

            while retries < max_retries:
                try:
                    span = "MAX"
                    data_headers = {
                        "authority": "prod-api.154310543964.hellopublic.com",
                        "method": "GET",
                        "path": f"/fixedincomegateway/v1/graph/data?cusip={cusip}&span={span}",
                        "scheme": "https",
                        "accept": "*/*",
                        "accept-encoding": "gzip, deflate, br, zstd",
                        "accept-language": "en-US,en;q=0.9",
                        "cache-control": "no-cache",
                        "content-type": "application/json",
                        "dnt": "1",
                        "origin": "https://public.com",
                        "pragma": "no-cache",
                        "priority": "u=1, i",
                        "sec-ch-ua": '"Chromium";v="128", "Not;A=Brand";v="24", "Google Chrome";v="128"',
                        "sec-ch-ua-mobile": "?0",
                        "sec-ch-ua-platform": '"Windows"',
                        "sec-fetch-dest": "empty",
                        "sec-fetch-mode": "cors",
                        "sec-fetch-site": "cross-site",
                        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36",
                        "x-app-version": "web-1.0.9",
                        "authorization": jwt_str,
                    }

                    data_url = f"https://prod-api.154310543964.hellopublic.com/fixedincomegateway/v1/graph/data?cusip={cusip}&span={span}"
                    response = await client.get(data_url, headers=data_headers)
                    response.raise_for_status()
                    df = pd.DataFrame(response.json()["data"])
                    df["timestamp"] = pd.to_datetime(df["timestamp"], errors="coerce")
                    df["unitPrice"] = pd.to_numeric(df["unitPrice"]) * 100
                    df["yieldToWorst"] = pd.to_numeric(df["yieldToWorst"]) * 100
                    df.columns = cols_to_return
                    if start_date:
                        df = df[df["Date"].dt.date >= start_date.date()]
                    if end_date:
                        df = df[df["Date"].dt.date <= end_date.date()]
                    if uid:
                        return cusip, df, uid
                    return cusip, df

                except httpx.HTTPStatusError as e:
                    self._logger.error(
                        f"Public.com - Bad Status for {cusip}: {response.status_code}"
                    )
                    if (
                        response.status_code == 404 or response.status_code == 400
                    ):  # public.com endpoint doesnt throw a 404 specifically
                        if uid:
                            return cusip, pd.DataFrame(columns=cols_to_return), uid
                        return cusip, pd.DataFrame(columns=cols_to_return)

                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(
                        f"Public.com - Throttled for {cusip}. Waiting for {wait_time} seconds before retrying..."
                    )
                    await asyncio.sleep(wait_time)

                except Exception as e:
                    self._logger.error(f"Public.com - Error: {str(e)}")
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(
                        f"Public.com - Throttled for {cusip}. Waiting for {wait_time} seconds before retrying..."
                    )
                    await asyncio.sleep(wait_time)

            raise ValueError(f"Public.com - Max retries exceeded for {cusip}")

        except Exception as e:
            self._logger.error(e)
            if uid:
                return cusip, pd.DataFrame(columns=cols_to_return), uid
            return cusip, pd.DataFrame(columns=cols_to_return)

    async def _fetch_cusip_timeseries_public_dotcome_with_semaphore(
        self, semaphore, *args, **kwargs
    ):
        async with semaphore:
            return await self._fetch_cusip_timeseries_public_dotcom(*args, **kwargs)

    async def _build_fetch_tasks_cusip_timeseries_public_dotcome(
        self,
        client: httpx.AsyncClient,
        cusips: List[str],
        start_date: datetime,
        end_date: datetime,
        uid: Optional[str | int] = None,
        max_concurrent_tasks: int = 64,
        refresh_jwt: Optional[bool] = False,
    ):
        if refresh_jwt or not self._public_dotcom_jwt:
            self._public_dotcom_jwt = self._fetch_public_dotcome_jwt()
            if not self._public_dotcom_jwt:
                raise ValueError("Public.com JWT Request Failed")

        semaphore = asyncio.Semaphore(max_concurrent_tasks)
        tasks = [
            self._fetch_cusip_timeseries_public_dotcome_with_semaphore(
                semaphore=semaphore,
                client=client,
                cusip=cusip,
                start_date=start_date,
                end_date=end_date,
                uid=uid,
                jwt_str=self._public_dotcom_jwt,
                max_retries=1,
            )
            for cusip in cusips
        ]
        return tasks

    def public_dotcom_timeseries_api(
        self,
        cusips: List[str],
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        refresh_jwt: Optional[bool] = False,
        max_concurrent_tasks: int = 64,
    ):
        if refresh_jwt or not self._public_dotcom_jwt:
            self._public_dotcom_jwt = self._fetch_public_dotcome_jwt()
            if not self._public_dotcom_jwt:
                raise ValueError("Public.com JWT Request Failed")

        async def build_tasks(
            client: httpx.AsyncClient,
            cusips: List[str],
            start_date: datetime,
            end_date: datetime,
            jwt_str: str,
        ):
            semaphore = asyncio.Semaphore(max_concurrent_tasks)
            tasks = [
                self._fetch_cusip_timeseries_public_dotcome_with_semaphore(
                    semaphore=semaphore,
                    client=client,
                    cusip=cusip,
                    start_date=start_date,
                    end_date=end_date,
                    jwt_str=jwt_str,
                    max_retries=1,
                )
                for cusip in cusips
            ]
            return await asyncio.gather(*tasks)

        async def run_fetch_all(
            cusips: List[str], start_date: datetime, end_date: datetime, jwt_str: str
        ):
            async with httpx.AsyncClient(proxy=self._proxies["https"]) as client:
                all_data = await build_tasks(
                    client=client,
                    cusips=cusips,
                    start_date=start_date,
                    end_date=end_date,
                    jwt_str=jwt_str,
                )
                return all_data

        dfs: List[Tuple[str, pd.DataFrame]] = asyncio.run(
            run_fetch_all(
                cusips=cusips,
                start_date=start_date,
                end_date=end_date,
                jwt_str=self._public_dotcom_jwt,
            )
        )
        return dict(dfs)



import asyncio
import multiprocessing as mp
import warnings
from datetime import datetime
from functools import partial, reduce
from typing import Dict, List, Optional, Tuple, Callable
import tqdm
import tqdm.asyncio

import httpx
import numpy as np
import pandas as pd
import polars as pl
from pandas.tseries.offsets import CustomBusinessDay, BDay
from pandas.tseries.holiday import USFederalHolidayCalendar
from scipy.optimize import newton

from CurvyCUSIPs.CurveInterpolator import GeneralCurveInterpolator
from CurvyCUSIPs.DataFetcher.bondsupermart import BondSupermartDataFetcher
from CurvyCUSIPs.DataFetcher.fedinvest import FedInvestDataFetcher
from CurvyCUSIPs.DataFetcher.finra import FinraDataFetcher
from CurvyCUSIPs.DataFetcher.fred import FredDataFetcher
from CurvyCUSIPs.DataFetcher.nyfrb import NYFRBDataFetcher
from CurvyCUSIPs.DataFetcher.public_dotcom import PublicDotcomDataFetcher
from CurvyCUSIPs.DataFetcher.ust import USTreasuryDataFetcher
from CurvyCUSIPs.DataFetcher.wsj import WSJDataFetcher
from CurvyCUSIPs.DataFetcher.yf import YahooFinanceDataFetcher
from CurvyCUSIPs.DataFetcher.bbg_sef import BBGSEF_DataFetcher
from CurvyCUSIPs.utils.QL_BondPricer import QL_BondPricer
from CurvyCUSIPs.utils.RL_BondPricer import RL_BondPricer
from CurvyCUSIPs.utils.ust_utils import get_active_cusips, get_last_n_off_the_run_cusips, is_valid_ust_cusip, ust_sorter, NoneReturningSpline

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)


def calculate_yields(row, as_of_date, use_quantlib=False):
    if use_quantlib:
        offer_yield = QL_BondPricer.bond_price_to_ytm(
            type=row["security_type"],
            issue_date=row["issue_date"],
            maturity_date=row["maturity_date"],
            as_of=as_of_date,
            coupon=row["int_rate"] / 100,
            price=row["offer_price"],
        )
        bid_yield = QL_BondPricer.bond_price_to_ytm(
            type=row["security_type"],
            issue_date=row["issue_date"],
            maturity_date=row["maturity_date"],
            as_of=as_of_date,
            coupon=row["int_rate"] / 100,
            price=row["bid_price"],
        )
        eod_yield = QL_BondPricer.bond_price_to_ytm(
            type=row["security_type"],
            issue_date=row["issue_date"],
            maturity_date=row["maturity_date"],
            as_of=as_of_date,
            coupon=row["int_rate"] / 100,
            price=row["eod_price"],
        )
    else:
        offer_yield = RL_BondPricer.bond_price_to_ytm(
            type=row["security_type"],
            issue_date=row["issue_date"],
            maturity_date=row["maturity_date"],
            as_of=as_of_date,
            coupon=row["int_rate"] / 100,
            price=row["offer_price"],
        )
        bid_yield = RL_BondPricer.bond_price_to_ytm(
            type=row["security_type"],
            issue_date=row["issue_date"],
            maturity_date=row["maturity_date"],
            as_of=as_of_date,
            coupon=row["int_rate"] / 100,
            price=row["bid_price"],
        )
        eod_yield = RL_BondPricer.bond_price_to_ytm(
            type=row["security_type"],
            issue_date=row["issue_date"],
            maturity_date=row["maturity_date"],
            as_of=as_of_date,
            coupon=row["int_rate"] / 100,
            price=row["eod_price"],
        )

    return offer_yield, bid_yield, eod_yield


class CurveDataFetcher:
    ust_data_fetcher: USTreasuryDataFetcher = None
    fedinvest_data_fetcher: FedInvestDataFetcher = None
    nyfrb_data_fetcher: NYFRBDataFetcher = None
    publicdotcom_data_fetcher: PublicDotcomDataFetcher = None
    fred_data_fetcher: FredDataFetcher = None
    finra_data_fetcher: FinraDataFetcher = None
    bondsupermart_fetcher: BondSupermartDataFetcher = None
    wsj_data_fetcher: WSJDataFetcher = None
    yf_data_fetcher: YahooFinanceDataFetcher = None

    def __init__(
        self,
        use_ust_issue_date: Optional[bool] = False,
        fred_api_key: Optional[str] = None,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        self.ust_data_fetcher = USTreasuryDataFetcher(
            use_ust_issue_date=use_ust_issue_date,
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self.fedinvest_data_fetcher = FedInvestDataFetcher(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self.nyfrb_data_fetcher = NYFRBDataFetcher(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self.publicdotcom_data_fetcher = PublicDotcomDataFetcher(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self.fred_data_fetcher = FredDataFetcher(
            fred_api_key=fred_api_key,
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self.finra_data_fetcher = FinraDataFetcher(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self.bondsupermart_fetcher = BondSupermartDataFetcher(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self.wsj_data_fetcher = WSJDataFetcher(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

        self.yf_data_fetcher = YahooFinanceDataFetcher(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )
        
        self.bbg_sef_fetcher = BBGSEF_DataFetcher(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        ) 

    @staticmethod
    def par_bond_equation(c, maturity, zero_curve_func):
        discounted_cash_flows = sum((c / 2) * np.exp(-(zero_curve_func(t) / 100) * t) for t in np.arange(0.5, maturity + 0.5, 0.5))
        final_payment = 100 * np.exp(-(zero_curve_func(maturity) / 100) * maturity)
        return discounted_cash_flows + final_payment - 100

    @staticmethod
    def par_curve_func(tenor, zero_curve_func):
        init_guess = 4
        return newton(
            CurveDataFetcher.par_bond_equation,
            x0=init_guess,
            args=(tenor, zero_curve_func),
        )

    def async_runner(self, tasks):
        pass

    # one of cme_ust_labels, ust_label_spread, cusip_spread has to be defined
    # default delimtter "/"
    def fetch_spreads(
        self,
        start_date: datetime,
        end_date: datetime,
        use_bid_side: Optional[bool] = False,
        use_offer_side: Optional[bool] = False,
        use_mid_side: Optional[bool] = False,
        cme_ust_labels: Optional[str] = None,
        ust_label_spread: Optional[str] = None,
        cusip_spread: Optional[str] = None,
        spread_delimtter: Optional[str] = "/",
    ) -> pd.DataFrame:
        spread_label = ""
        if ust_label_spread:
            labels_to_fetch = ust_label_spread.split(spread_delimtter)
            cusips_to_fetch = [self.ust_data_fetcher.ust_label_to_cusip(label.strip())["cusip"] for label in labels_to_fetch]
            spread_label = ust_label_spread
        if cme_ust_labels:
            labels_to_fetch = cme_ust_labels.split(spread_delimtter)
            cusips_to_fetch = [self.ust_data_fetcher.cme_ust_label_to_cusip(label.strip())["cusip"] for label in labels_to_fetch]
            spread_label = cme_ust_labels
        if cusip_spread:
            cusips_to_fetch = cusip_spread.split(spread_delimtter)
            spread_label = cusip_spread

        if len(cusips_to_fetch) < 2:
            return "not a valid spread"

        cusip_dict_df = self.fedinvest_data_fetcher.cusips_timeseries(cusips=cusips_to_fetch, start_date=start_date, end_date=end_date)

        yield_col = "eod_yield"
        if use_bid_side:
            yield_col = "bid_yield"
        elif use_offer_side:
            yield_col = "offer_yield"
        elif use_mid_side:
            yield_col = "mid_yield"

        dfs = [
            df[["Date", yield_col]].rename(columns={yield_col: f"{self.ust_data_fetcher.cusip_to_ust_label(key)}"})
            for key, df in cusip_dict_df.items()
        ]
        merged_df = reduce(lambda left, right: pd.merge(left, right, on="Date", how="outer"), dfs)
        if len(cusips_to_fetch) == 3:
            merged_df[spread_label] = (merged_df.iloc[:, 2] - merged_df.iloc[:, 1]) - ((merged_df.iloc[:, 3] - merged_df.iloc[:, 2]))
        else:
            merged_df[spread_label] = merged_df.iloc[:, 2] - merged_df.iloc[:, 1]

        return merged_df

    # max_concurrent_tasks limit applies to the logical tasks that your code wants to execute concurrently
    #    - ensures that we don’t attempt to start too many logical fetching operations at once
    # max_connections limit applies to the physical HTTP requests that can be made concurrently by httpx.AsyncClient
    #    - ensures only a certain number can actually initiate network connections concurrently
    # fitted_curves is tuple of size 3 (interp key, quote yield type, callable curve set df filter function)
    def fetch_historical_curve_sets(
        self,
        start_date: datetime,
        end_date: Optional[datetime] = None,
        fetch_soma_holdings: Optional[bool] = False,
        fetch_stripping_data: Optional[bool] = False,
        calc_free_float: Optional[bool] = False,
        fitted_curves: Optional[List[Tuple[str, str, Callable]] | List[Tuple[str, str, Callable, Callable]]] = None,
        max_concurrent_tasks: Optional[int] = 128,
        max_connections: Optional[int] = 64,
        sorted_curve_set: Optional[bool] = False, 
    ) -> Tuple[Dict[datetime, pd.DataFrame], Dict[datetime, Dict[str, GeneralCurveInterpolator]]]:
        if not end_date:
            end_date = start_date
            
        if calc_free_float:
            fetch_soma_holdings = True
            fetch_stripping_data = True

        async def gather_tasks(client: httpx.AsyncClient, dates: datetime, max_concurrent_tasks):
            my_semaphore = asyncio.Semaphore(max_concurrent_tasks)
            tasks = await self.fedinvest_data_fetcher._build_fetch_tasks_cusip_prices_from_github(
                client=client,
                dates=dates,
                uid="ust_prices",
                cols_to_return=(
                    [
                        "cusip",
                        "offer_price",
                        "offer_yield",
                        "bid_price",
                        "bid_yield",
                        "mid_price",
                        "mid_yield",
                        "eod_price",
                        "eod_yield",
                    ]
                ),
                my_semaphore=my_semaphore,
            )

            if fetch_soma_holdings:
                soma_bwd_date: pd.Timestamp = start_date - BDay(5)
                tasks += await self.nyfrb_data_fetcher._build_fetch_tasks_historical_soma_holdings(
                    client=client,
                    dates=[soma_bwd_date.to_pydatetime()] + dates,
                    uid="soma_holdings",
                    minimize_api_calls=True,
                    my_semaphore=my_semaphore,
                )

            if fetch_stripping_data:
                strips_bwd_date: pd.Timestamp = start_date - BDay(20)
                tasks += await self.ust_data_fetcher._build_fetch_tasks_historical_stripping_activity(
                    client=client,
                    dates=[strips_bwd_date.to_pydatetime()] + dates,
                    uid="ust_stripping",
                    minimize_api_calls=True,
                    my_semaphore=my_semaphore,
                )

            # return await asyncio.gather(*tasks)
            return await tqdm.asyncio.tqdm.gather(*tasks, desc="FETCHING CURVE SETS...")

        async def run_fetch_all(dates: datetime, max_concurrent_tasks: int, max_connections: int):
            limits = httpx.Limits(max_connections=max_connections)
            async with httpx.AsyncClient(
                limits=limits,
            ) as client:
                all_data = await gather_tasks(client=client, dates=dates, max_concurrent_tasks=max_concurrent_tasks)
                return all_data

        bdates = [
            bday.to_pydatetime()
            for bday in pd.bdate_range(
                start=start_date,
                end=end_date,
                freq=CustomBusinessDay(calendar=USFederalHolidayCalendar()),
            )
        ]

        results: List[Tuple[datetime, pd.DataFrame, str]] = asyncio.run(
            run_fetch_all(dates=bdates, max_concurrent_tasks=max_concurrent_tasks, max_connections=max_connections)
        )
        sorted_results = sorted(results, key=lambda x: x[0])

        auctions_df: pl.DataFrame = pl.from_pandas(self.ust_data_fetcher._historical_auctions_df.copy())
        auctions_df = auctions_df.filter(
            (pl.col("security_type") == "Bill") | (pl.col("security_type") == "Note") | (pl.col("security_type") == "Bond")
        )
        auctions_df = auctions_df.with_columns(
            pl.when(pl.col("original_security_term").str.contains("29-Year"))
            .then(pl.lit("30-Year"))
            .when(pl.col("original_security_term").str.contains("30-"))
            .then(pl.lit("30-Year"))
            .otherwise(pl.col("original_security_term"))
            .alias("original_security_term")
        )

        last_seen_soma_holdings_df = None
        last_seen_stripping_act_df = None
        curveset_dict_df: Dict[datetime, List[pd.DataFrame]] = {}
        curveset_intrep_dict: Dict[datetime, Dict[str, GeneralCurveInterpolator]] = {}

        for tup in tqdm.tqdm(sorted_results, desc="AGGREGATING CURVE SET DFs"):
            curr_dt = tup[0]
            uid = tup[-1]

            if tup[1] is None or tup[1].empty:
                continue

            if uid == "soma_holdings":
                last_seen_soma_holdings_df = pl.from_pandas(tup[1])
                continue

            if uid == "ust_stripping":
                last_seen_stripping_act_df = pl.from_pandas(tup[1])
                continue

            if uid == "ust_prices":
                price_df = pl.from_pandas(tup[1])
                curr_auctions_df = auctions_df.filter(
                    (pl.col("issue_date").dt.date() <= curr_dt.date()) & (pl.col("maturity_date") >= curr_dt)
                ).unique(subset=["cusip"], keep="first")

                merged_df = curr_auctions_df.join(price_df, on="cusip", how="outer")

                if fetch_soma_holdings and last_seen_soma_holdings_df is not None:
                    merged_df = merged_df.join(last_seen_soma_holdings_df, on="cusip", how="left")
                if fetch_stripping_data and last_seen_stripping_act_df is not None:
                    merged_df = merged_df.join(last_seen_stripping_act_df, on="cusip", how="left")

                merged_df = merged_df.filter(pl.col("cusip").map_elements(is_valid_ust_cusip, return_dtype=pl.Boolean))
                merged_df = merged_df.with_columns(pl.col("maturity_date").cast(pl.Datetime).alias("maturity_date"))
                merged_df = merged_df.with_columns(((pl.col("maturity_date") - curr_dt).dt.total_days() / 365).alias("time_to_maturity"))
                merged_df = merged_df.with_columns(
                    pl.col("time_to_maturity").rank(descending=True, method="ordinal").over("original_security_term").sub(1).alias("rank")
                )

                if calc_free_float:
                    merged_df = merged_df.with_columns(
                        pl.col("parValue").cast(pl.Float64).fill_null(0).alias("parValue"),
                        (pl.col("portion_stripped_amt").cast(pl.Float64).fill_null(0) * 1000).alias("portion_stripped_amt"),
                        (
                            pl.when((pl.col("est_outstanding_amt").is_not_nan()) & (pl.col("est_outstanding_amt") != 0))
                            .then(pl.col("est_outstanding_amt"))
                            .otherwise(pl.col("outstanding_amt"))
                            .cast(pl.Float64)
                            .fill_null(0)
                            * 1000
                        ).alias("est_outstanding_amt"),
                    )
                    merged_df = merged_df.with_columns(
                        ((pl.col("est_outstanding_amt") - pl.col("parValue") - pl.col("portion_stripped_amt")) / 1_000_000).alias("free_float")
                    )

                curr_curve_set_df = merged_df.to_pandas()
                if sorted_curve_set:
                    curr_curve_set_df["sort_key"] = curr_curve_set_df["original_security_term"].apply(ust_sorter)
                    curr_curve_set_df = curr_curve_set_df.sort_values(by=["sort_key", "time_to_maturity"]).drop(columns="sort_key").reset_index(drop=True)
                
                curveset_dict_df[curr_dt] = curr_curve_set_df

                if fitted_curves:
                    for curve_build_params in fitted_curves:
                        if len(curve_build_params) == 3:
                            if callable(curve_build_params[-1]):
                                curve_set_key, quote_type, filter_func = curve_build_params
                                curr_filtered_curve_set_df: pd.DataFrame = filter_func(curr_curve_set_df)

                                if curr_dt not in curveset_intrep_dict:
                                    curveset_intrep_dict[curr_dt] = {}
                                if curve_set_key not in curveset_intrep_dict[curr_dt]:
                                    curveset_intrep_dict[curr_dt][curve_set_key] = {}

                                curveset_intrep_dict[curr_dt][curve_set_key] = GeneralCurveInterpolator(
                                    x=curr_filtered_curve_set_df["time_to_maturity"].to_numpy(), y=curr_filtered_curve_set_df[quote_type].to_numpy()
                                )

                        elif len(curve_build_params) == 4:
                            if callable(curve_build_params[-2]) and callable(curve_build_params[-2]):
                                curve_set_key, quote_type, filter_func, calibrate_func = curve_build_params
                                curr_filtered_curve_set_df: pd.DataFrame = filter_func(curr_curve_set_df)

                                if curr_dt not in curveset_intrep_dict:
                                    curveset_intrep_dict[curr_dt] = {}
                                if curve_set_key not in curveset_intrep_dict[curr_dt]:
                                    curveset_intrep_dict[curr_dt][curve_set_key] = {}

                                try:
                                    curr_filtered_curve_set_df = (
                                        curr_filtered_curve_set_df[["time_to_maturity", quote_type]].dropna().sort_values(by="time_to_maturity")
                                    )
                                    parameteric_model = calibrate_func(
                                        curr_filtered_curve_set_df["time_to_maturity"].to_numpy(),
                                        curr_filtered_curve_set_df[quote_type].to_numpy(),
                                    )
                                    assert parameteric_model[1]
                                    curveset_intrep_dict[curr_dt][curve_set_key] = parameteric_model[0]

                                except Exception as e:
                                    # print(f"{curve_set_key} for {curr_dt} - {str(e)}")
                                    curveset_intrep_dict[curr_dt][curve_set_key] = NoneReturningSpline(
                                        curr_filtered_curve_set_df["time_to_maturity"].to_numpy(),
                                        curr_filtered_curve_set_df[quote_type].to_numpy(),
                                    )

        if fitted_curves:
            return curveset_dict_df, curveset_intrep_dict

        return curveset_dict_df

    def build_curve_set(
        self,
        as_of_date: datetime,
        calc_ytms: Optional[bool] = True,
        use_quantlib: Optional[bool] = False,  # default is rateslib
        include_auction_results: Optional[bool] = False,
        include_soma_holdings: Optional[bool] = False,
        include_stripping_activity: Optional[bool] = False,
        # include_outstanding_amt: Optional[bool] = False,
        # exclude_nonmarketable_outstanding_amt: Optional[bool] = False,
        # auctions_df: Optional[pd.DataFrame] = None,
        sorted: Optional[bool] = False,
        use_github: Optional[bool] = False,
        use_public_dotcom: Optional[bool] = False,
        include_off_the_run_number: Optional[bool] = False,
        include_corpus_cusip: Optional[bool] = False,
        market_cols_to_return: List[str] = None,
        calc_free_float: Optional[bool] = False,
        calc_mod_duration: Optional[bool] = False,
    ):
        if as_of_date.date() > datetime.today().date():
            print(f"crystal ball feature not implemented, yet - {as_of_date} is in the future")
            return

        if use_github or use_public_dotcom:
            calc_ytms = False

        if market_cols_to_return:
            if "cusip" not in market_cols_to_return:
                market_cols_to_return.insert(0, "cusip")

        quote_type = market_cols_to_return[1].split("_")[0] if market_cols_to_return else "eod"
        filtered_free_float_df_col = False
        if calc_free_float:
            if not include_soma_holdings and not include_auction_results:
                filtered_free_float_df_col = True
                include_auction_results = True
                include_soma_holdings = True
                include_stripping_activity = True
                # include_outstanding_amt = True

        async def gather_tasks(client: httpx.AsyncClient, as_of_date: datetime):
            if use_github:
                ust_historical_prices_tasks = await self.fedinvest_data_fetcher._build_fetch_tasks_cusip_prices_from_github(
                    client=client,
                    dates=[as_of_date],
                    uid="ust_prices",
                    cols_to_return=(
                        [
                            "cusip",
                            "offer_price",
                            "offer_yield",
                            "bid_price",
                            "bid_yield",
                            "mid_price",
                            "mid_yield",
                            "eod_price",
                            "eod_yield",
                        ]
                        if not market_cols_to_return
                        else market_cols_to_return
                    ),
                )
            elif use_public_dotcom:
                cusips_to_fetch = get_active_cusips(
                    historical_auctions_df=self.ust_data_fetcher._historical_auctions_df,
                    as_of_date=as_of_date,
                    use_issue_date=True,
                )["cusip"].to_list()
                ust_historical_prices_tasks = await self.publicdotcom_data_fetcher._build_fetch_tasks_cusip_timeseries_public_dotcome(
                    client=client,
                    cusips=cusips_to_fetch,
                    start_date=as_of_date,
                    end_date=as_of_date,
                    uid="ust_prices_public_dotcom",
                )
            else:
                ust_historical_prices_tasks = await self.fedinvest_data_fetcher._build_fetch_tasks_cusip_prices_fedinvest(
                    client=client, dates=[as_of_date], uid="ust_prices"
                )

            tasks = ust_historical_prices_tasks

            if include_soma_holdings:
                tasks += await self.nyfrb_data_fetcher._build_fetch_tasks_historical_soma_holdings(
                    client=client, dates=[as_of_date], uid="soma_holdings"
                )
            if include_stripping_activity:
                tasks += await self.ust_data_fetcher._build_fetch_tasks_historical_stripping_activity(
                    client=client, dates=[as_of_date], uid="ust_stripping"
                )
            # if include_outstanding_amt:
            #     tasks += await self._build_fetch_tasks_historical_amount_outstanding(
            #         client=client,
            #         dates=[as_of_date],
            #         uid="ust_outstanding_amt",
            #         only_marketable=exclude_nonmarketable_outstanding_amt,
            #     )

            return await asyncio.gather(*tasks)

        async def run_fetch_all(as_of_date: datetime):
            limits = httpx.Limits(max_connections=10)
            async with httpx.AsyncClient(
                limits=limits,
            ) as client:
                all_data = await gather_tasks(client=client, as_of_date=as_of_date)
                return all_data

        results = asyncio.run(run_fetch_all(as_of_date=as_of_date))
        auctions_dfs = []
        dfs = []
        public_dotcom_dicts: List[Dict[str, str]] = []
        for tup in results:
            uid = tup[-1]
            if uid == "ust_auctions":
                auctions_dfs.append(tup[0])
                continue

            if uid == "ust_prices" or uid == "soma_holdings" or uid == "ust_stripping" or uid == "ust_outstanding_amt":
                dfs.append(tup[1])
                continue

            if uid == "ust_prices_public_dotcom" and use_public_dotcom:
                if not isinstance(tup[1], pd.DataFrame):
                    continue
                if not tup[1].empty:
                    public_dotcom_dicts.append(
                        {
                            "cusip": tup[0],
                            f"{quote_type}_price": tup[1].iloc[-1]["Price"],
                            f"{quote_type}_yield": tup[1].iloc[-1]["YTM"],
                        }
                    )
                continue

            # should neever get here
            print(f"CURVE SET - unknown UID, Current Tuple: {tup}")

        auctions_df = get_active_cusips(
            historical_auctions_df=self.ust_data_fetcher._historical_auctions_df,
            as_of_date=as_of_date,
            use_issue_date=True,
        )
        otr_cusips_df: pd.DataFrame = get_last_n_off_the_run_cusips(
            auctions_df=auctions_df,
            n=0,
            filtered=True,
            as_of_date=as_of_date,
            use_issue_date=self.ust_data_fetcher._use_ust_issue_date,
        )
        auctions_df["is_on_the_run"] = auctions_df["cusip"].isin(otr_cusips_df["cusip"].to_list())
        auctions_df["time_to_maturity"] = (auctions_df["maturity_date"] - as_of_date).dt.days / 365

        if not include_auction_results or filtered_free_float_df_col:
            default_auction_cols = [
                "cusip",
                "security_type",
                "auction_date",
                "issue_date",
                "maturity_date",
                "time_to_maturity",
                "int_rate",
                "high_investment_rate",
                "is_on_the_run",
                "ust_label",
                "security_term",
                "original_security_term",
            ]
            if include_corpus_cusip:
                default_auction_cols.append("corpus_cusip")
            auctions_df = auctions_df[default_auction_cols]

        merged_df = reduce(lambda left, right: pd.merge(left, right, on="cusip", how="outer"), dfs)
        if not use_public_dotcom:
            merged_df = pd.merge(left=auctions_df, right=merged_df, on="cusip", how="outer")
        else:
            market_df = pd.merge(
                left=auctions_df,
                right=pd.DataFrame(public_dotcom_dicts),
                on="cusip",
                how="outer",
            )
            merged_df = pd.merge(
                left=market_df,
                right=merged_df,
                on="cusip",
                how="outer",
            )
        merged_df = merged_df[merged_df["cusip"].apply(is_valid_ust_cusip)]

        if calc_free_float:
            merged_df["parValue"] = pd.to_numeric(merged_df["parValue"], errors="coerce").fillna(0)
            merged_df["portion_stripped_amt"] = pd.to_numeric(merged_df["portion_stripped_amt"], errors="coerce").fillna(0) * 1000
            merged_df["outstanding_amt"] = pd.to_numeric(merged_df["outstanding_amt"], errors="coerce").fillna(0) * 1000
            merged_df["free_float"] = (merged_df["outstanding_amt"] - merged_df["parValue"] - merged_df["portion_stripped_amt"]) / 1_000_000

        if calc_mod_duration:
            merged_df["mod_dur"] = merged_df.apply(
                lambda row: RL_BondPricer._bond_mod_duration(
                    issue_date=row["issue_date"],
                    maturity_date=row["maturity_date"],
                    as_of=as_of_date,
                    coupon=row["int_rate"] / 100,
                    ytm=(
                        row[
                            next(
                                (item for item in market_cols_to_return if "yield" in item),
                                None,
                            )
                        ]
                        if market_cols_to_return
                        else row["eod_yield"]
                    ),
                ),
                axis=1,
            )

        if not market_cols_to_return:
            if not use_github:
                merged_df["mid_price"] = (merged_df["offer_price"] + merged_df["bid_price"]) / 2
            else:
                merged_df["mid_yield"] = (merged_df["offer_yield"] + merged_df["bid_yield"]) / 2

        if calc_ytms:
            calculate_yields_partial = partial(calculate_yields, as_of_date=as_of_date, use_quantlib=use_quantlib)
            with mp.Pool(mp.cpu_count()) as pool:
                results = pool.map(calculate_yields_partial, [row for _, row in merged_df.iterrows()])
            offer_yields, bid_yields, eod_yields = zip(*results)
            merged_df["offer_yield"] = offer_yields
            merged_df["bid_yield"] = bid_yields
            merged_df["eod_yield"] = eod_yields
            merged_df["mid_yield"] = (merged_df["offer_yield"] + merged_df["bid_yield"]) / 2

        merged_df = merged_df.replace("null", np.nan)
        merged_df = merged_df[merged_df["original_security_term"].notna()]
        if sorted:
            merged_df["sort_key"] = merged_df["original_security_term"].apply(ust_sorter)
            merged_df = merged_df.sort_values(by=["sort_key", "time_to_maturity"]).drop(columns="sort_key").reset_index(drop=True)

        if include_off_the_run_number:
            merged_df["rank"] = merged_df.groupby("original_security_term")["time_to_maturity"].rank(ascending=False, method="first") - 1

        return merged_df



# reference:
# https://publications.gc.ca/site/archivee-archived.html?url=https://publications.gc.ca/collections/collection_2014/banque-bank-canada/FB3-2-102-29-eng.pdf


import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from dataclasses import dataclass


"""Implementation of the Merrill Lynch Exponential Spline interest rate curve model."""


@dataclass
class MerrillLynchExponentialSplineModel:

    alpha: float
    N: int
    lambda_hat: np.ndarray

    def e_basis(self, k: int, T: float) -> float:
        return np.exp(-self.alpha * k * T)

    def discount(self, T: float) -> float:
        return sum(self.lambda_hat[k] * self.e_basis(k, T) for k in range(self.N))

    def construct_H(self, T: np.ndarray) -> np.ndarray:
        H = np.zeros((len(T), self.N))
        for j in range(len(T)):
            for k in range(self.N):
                H[j, k] = self.e_basis(k, T[j])
        return H

    def gls(self, H: np.ndarray, W: np.ndarray, p: np.ndarray) -> np.ndarray:
        """Generalized Least Squares (GLS) estimate of the MLES basis parameters."""
        return np.linalg.inv(H.T @ W @ H) @ H.T @ W @ p

    def fit(self, maturities: np.ndarray, yields: np.ndarray, W: np.ndarray):
        """Fits the model to observed yields."""
        prices = np.exp(-yields * maturities / 100)
        H = self.construct_H(maturities)
        # Perform GLS estimation
        self.lambda_hat = self.gls(H, W, prices)

    def theoretical_yields(self, maturities: np.ndarray) -> np.ndarray:
        prices = np.array([self.discount(T) for T in maturities])
        return -np.log(prices) / maturities * 100

    def __call__(self, maturities: np.ndarray) -> np.ndarray:
        return self.theoretical_yields(maturities)



from datetime import datetime
from typing import Literal

import numpy as np
import pandas as pd
import QuantLib as ql


class QL_BondPricer:

    @staticmethod
    def _pydatetime_to_qldate(date: datetime) -> ql.Date:
        return ql.Date(date.day, date.month, date.year)

    @staticmethod
    def _bill_price_to_ytm(maturity_date: datetime, as_of: datetime, price: float) -> float:
        settlement_date = (pd.to_datetime(as_of) + pd.tseries.offsets.BDay(0)).to_pydatetime()
        t = (maturity_date - settlement_date).days
        if t <= 0:
            return np.nan

        F = 100.0
        HPY = (F - price) / price
        BEY = HPY * (365 / t)
        return BEY * 100

    @staticmethod
    def _coupon_bond_price_to_ytm(
        issue_date: datetime,
        maturity_date: datetime,
        as_of: datetime,
        coupon: float,
        price: float,
    ) -> float:
        issue_ql_date = QL_BondPricer._pydatetime_to_qldate(issue_date)
        maturity_ql_date = QL_BondPricer._pydatetime_to_qldate(maturity_date)
        settlement_date = (pd.to_datetime(as_of) + pd.tseries.offsets.BDay(1)).to_pydatetime()
        settlement_ql_date = QL_BondPricer._pydatetime_to_qldate(settlement_date)

        clean_price = price
        coupon_rate = coupon / 100
        day_count = ql.ActualActual(ql.ActualActual.Bond)
        coupon_frequency = ql.Semiannual
        schedule = ql.Schedule(
            issue_ql_date,
            maturity_ql_date,
            ql.Period(coupon_frequency),
            ql.UnitedStates(ql.UnitedStates.GovernmentBond),
            ql.Unadjusted,
            ql.Unadjusted,
            ql.DateGeneration.Backward,
            False,
        )

        settlement_days = 1
        bond = ql.FixedRateBond(settlement_days, 100.0, schedule, [coupon_rate], day_count)
        bond_price_handle = ql.QuoteHandle(ql.SimpleQuote(clean_price))
        bond_engine = ql.DiscountingBondEngine(ql.YieldTermStructureHandle(ql.FlatForward(settlement_ql_date, 0.0, day_count)))
        bond.setPricingEngine(bond_engine)
        ytm = bond.bondYield(
            bond_price_handle.currentLink().value(),
            day_count,
            ql.Compounded,
            coupon_frequency,
        )
        return ytm * 100

    @staticmethod
    def bond_price_to_ytm(
        type: Literal["Bill", "Note", "Bond"],
        issue_date: datetime,
        maturity_date: datetime,
        as_of: datetime,
        coupon: float,
        price: float,
    ):
        try:
            if type == "Bill":
                return QL_BondPricer._bill_price_to_ytm(
                    maturity_date=maturity_date,
                    as_of=as_of,
                    price=price,
                )

            return QL_BondPricer._coupon_bond_price_to_ytm(
                issue_date=issue_date,
                maturity_date=maturity_date,
                as_of=as_of,
                coupon=coupon,
                price=price,
            )
        except:
            return np.nan



from datetime import datetime
from typing import Literal

import numpy as np
import pandas as pd
import rateslib as rl
from pandas.tseries.offsets import BDay
from typing import Optional


class RL_BondPricer:

    @staticmethod
    def _pydatetime_to_rldate(date: datetime) -> rl.dt:
        return rl.dt(date.year, date.month, date.day)

    @staticmethod
    def _bill_price_to_ytm(
        issue_date: datetime, maturity_date: datetime, as_of: datetime, price: float
    ) -> float:
        bill_ust = rl.Bill(
            effective=RL_BondPricer._pydatetime_to_rldate(issue_date),
            termination=RL_BondPricer._pydatetime_to_rldate(maturity_date),
            calendar="nyc",
            modifier="NONE",
            currency="usd",
            convention="Act360",
            settle=1,
            notional=-1e6,  # negative notional receives fixed, i.e. buys a bill
            curves="bill_curve",
            calc_mode="us_gbb",
        )
        settle_pd_ts: pd.Timestamp = as_of + BDay(1)
        return bill_ust.ytm(price=price, settlement=settle_pd_ts.to_pydatetime())

    @staticmethod
    def _coupon_bond_price_to_ytm(
        issue_date: datetime,
        maturity_date: datetime,
        as_of: datetime,
        coupon: float,
        price: float,
    ) -> float:
        fxb_ust = rl.FixedRateBond(
            effective=RL_BondPricer._pydatetime_to_rldate(issue_date),
            termination=RL_BondPricer._pydatetime_to_rldate(maturity_date),
            fixed_rate=coupon * 100,
            spec="ust",
            calc_mode="ust_31bii",
        )
        settle_pd_ts: pd.Timestamp = as_of + BDay(1)
        return fxb_ust.ytm(price=price, settlement=settle_pd_ts.to_pydatetime())

    @staticmethod
    def _coupon_bond_ytm_to_price(
        issue_date: datetime,
        maturity_date: datetime,
        as_of: datetime,
        coupon: float,
        ytm: float,
        dirty: Optional[bool] = False,
    ) -> float:
        fxb_ust = rl.FixedRateBond(
            effective=RL_BondPricer._pydatetime_to_rldate(issue_date),
            termination=RL_BondPricer._pydatetime_to_rldate(maturity_date),
            fixed_rate=coupon * 100,
            spec="ust",
            calc_mode="ust_31bii",
        )
        settle_pd_ts: pd.Timestamp = as_of + BDay(1)
        return fxb_ust.price(
            ytm=ytm, settlement=settle_pd_ts.to_pydatetime(), dirty=dirty
        )

    @staticmethod
    def _bond_mod_duration(
        issue_date: datetime,
        maturity_date: datetime,
        as_of: datetime,
        coupon: float,
        ytm: float,
    ) -> float:
        try:
            fxb_ust = rl.FixedRateBond(
                effective=RL_BondPricer._pydatetime_to_rldate(issue_date),
                termination=RL_BondPricer._pydatetime_to_rldate(maturity_date),
                fixed_rate=coupon * 100,
                spec="ust",
            )
            settle_pd_ts: pd.Timestamp = as_of + BDay(1)
            return fxb_ust.duration(settlement=settle_pd_ts, ytm=ytm, metric="modified")
        except:
            return None

    @staticmethod
    def bond_price_to_ytm(
        type: Literal["Bill", "Note", "Bond"],
        issue_date: datetime,
        maturity_date: datetime,
        as_of: datetime,
        coupon: float,
        price: float,
    ):
        try:
            if type == "Bill":
                return RL_BondPricer._bill_price_to_ytm(
                    issue_date=issue_date,
                    maturity_date=maturity_date,
                    as_of=as_of,
                    price=price,
                )

            return RL_BondPricer._coupon_bond_price_to_ytm(
                issue_date=issue_date,
                maturity_date=maturity_date,
                as_of=as_of,
                coupon=coupon,
                price=price,
            )
        except Exception as e:
            # print(e)
            return np.nan



"""
interpolations:
- Cosine
- Chebyshev 
- Orthogonal polynomials
- Monomial interpolation
- Vasicek
- Fong-Vasicek (implemented MLESM which was inspired by )
- Cox-Ingersoll-Ross 
- Smith-Wilson
- gaussian process regression interpolation
- Gaussian Mixture Models (GMM)
- Principal Component Regression (PCR)
- Nearest neighbour weighted interpolation

"""

import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Callable, Dict, List, Literal, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.interpolate

from CurvyCUSIPs.models.MonotoneConvex import MonotoneConvex


# TODO
# explore numba
class GeneralCurveInterpolator:
    _x: npt.ArrayLike = None
    _y: npt.ArrayLike = None
    _linspace_x_num: int = 100
    _linspace_x: npt.NDArray[np.float64] = np.zeros(
        shape=_linspace_x_num, dtype=np.float64
    )

    _enable_extrapolate_left_fill: bool = False
    _enable_extrapolate_right_fill: bool = False

    _cubic_spline_interp_bc_type: Literal[
        "not-a-knot", "periodic", "clamped", "natural"
    ] = "not-a-knot"
    _cubic_spline_interp_custom_knots: npt.NDArray[np.float64] = np.array([])
    _cubic_spline_interp_nu: Literal[0, 1, 2] = 0

    _bspline_k: int = 3

    _logger = logging.getLogger()
    _debug_verbose: bool = False
    _info_verbose: bool = False  # performance benchmarking mainly
    _no_logs_plz: bool = False

    def __init__(
        self,
        x: npt.ArrayLike,
        y: npt.ArrayLike,
        linspace_x_lower_bound: Optional[float | int] = None,
        linspace_x_upper_bound: Optional[float | int] = None,
        linspace_x_num: int = 100,
        enable_extrapolate_left_fill: Optional[bool] = False,
        enable_extrapolate_right_fill: Optional[bool] = False,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        no_logs_plz: Optional[bool] = False,
    ):
        self._x = x
        self._y = y
        self._linspace_x_num = linspace_x_num
        self._linspace_x = np.linspace(
            linspace_x_lower_bound or min(x),
            linspace_x_upper_bound or max(x),
            num=self._linspace_x_num,
        )
        self._enable_extrapolate_left_fill = enable_extrapolate_left_fill
        self._enable_extrapolate_right_fill = enable_extrapolate_right_fill

        self._debug_verbose = debug_verbose
        self._info_verbose = info_verbose
        self._no_logs_plz = no_logs_plz
        if self._debug_verbose:
            self._logger.setLevel(self._logger.DEBUG)
        if self._info_verbose:
            self._logger.setLevel(self._logger.INFO)
        if self._no_logs_plz:
            self._logger.disabled = True
            self._logger.propagate = False

    def linear_interpolation(
        self, return_func=False
    ) -> np.ndarray | scipy.interpolate.interp1d:
        func_no_extrap = scipy.interpolate.interp1d(
            self._x, self._y, kind="linear", bounds_error=False
        )
        func_extrap = scipy.interpolate.interp1d(
            self._x,
            self._y,
            axis=0,
            kind="linear",
            bounds_error=False,
            fill_value="extrapolate",
        )
        if return_func:
            return func_extrap

        ynew_no_extrap = func_no_extrap(self._linspace_x)
        ynew_extrap = func_extrap(self._linspace_x)
        ynew = ynew_no_extrap.copy()
        if self._enable_extrapolate_left_fill:
            ynew[self._linspace_x < self._x[0]] = ynew_extrap[
                self._linspace_x < self._x[0]
            ]
        if self._enable_extrapolate_right_fill:
            ynew[self._linspace_x > self._x[-1]] = ynew_extrap[
                self._linspace_x > self._x[-1]
            ]

        return ynew

    def log_linear_interpolation(
        self, return_func=False
    ) -> np.ndarray | scipy.interpolate.interp1d:
        log_x = np.log(self._x)
        log_linspace_x = np.log(self._linspace_x)

        func_no_extrap = scipy.interpolate.interp1d(
            log_x, self._y, kind="linear", bounds_error=False
        )
        func_extrap = scipy.interpolate.interp1d(
            log_x,
            self._y,
            axis=0,
            kind="linear",
            bounds_error=False,
            fill_value="extrapolate",
        )
        if return_func:
            return func_extrap

        ynew_no_extrap = func_no_extrap(log_linspace_x)
        ynew_extrap = func_extrap(log_linspace_x)
        ynew = ynew_no_extrap.copy()
        if self._enable_extrapolate_left_fill:
            ynew[self._linspace_x < self._x[0]] = ynew_extrap[
                self._linspace_x < self._x[0]
            ]
        if self._enable_extrapolate_right_fill:
            ynew[self._linspace_x > self._x[-1]] = ynew_extrap[
                self._linspace_x > self._x[-1]
            ]

        return ynew

    def set_cubic_spline_interpolation_bc_type(
        self,
        bc_type: Literal["not-a-knot", "periodic", "clamped", "natural"] = "not-a-knot",
    ):
        self._cubic_spline_interp_bc_type = bc_type

    def set_cubic_spline_interpolation_custom_knots(
        self,
        custom_knots: npt.NDArray[np.float64],
    ):
        self._cubic_spline_interp_custom_knots = custom_knots

    def set_cubic_spline_interpolation_nu(self, nu: int):
        self._cubic_spline_interp_nu = nu

    def cubic_spline_interpolation(
        self,
        bc_type: Optional[
            Literal["not-a-knot", "periodic", "clamped", "natural"]
        ] = None,
        custom_knots: Optional[npt.NDArray[np.float64]] = None,
        nu: Optional[int] = None,
        return_func=False,
    ) -> np.ndarray | scipy.interpolate.CubicSpline:
        if bc_type:
            self._cubic_spline_interp_bc_type = bc_type
        if custom_knots:
            self._cubic_spline_interp_custom_knots = custom_knots
        if nu:
            self._cubic_spline_interp_nu = nu
        all_x = np.sort(
            np.concatenate([self._x, self._cubic_spline_interp_custom_knots])
        )
        all_y = np.interp(all_x, self._x, self._y)
        spline = scipy.interpolate.CubicSpline(
            all_x,
            all_y,
            bc_type=self._cubic_spline_interp_bc_type,
            extrapolate=True,
        )
        if return_func:
            return spline
        
        ynew = spline(self._linspace_x, nu=self._cubic_spline_interp_nu)
        if self._enable_extrapolate_left_fill:
            left_indices = self._linspace_x < min(self._x)
            ynew[left_indices] = self._y[0] + (
                self._linspace_x[left_indices] - self._x[0]
            ) * (self._y[1] - self._y[0]) / (self._x[1] - self._x[0])
        if self._enable_extrapolate_right_fill:
            right_indices = self._linspace_x > max(self._x)
            ynew[right_indices] = self._y[-1] + (
                self._linspace_x[right_indices] - self._x[-1]
            ) * (self._y[-1] - self._y[-2]) / (self._x[-1] - self._x[-2])

        return ynew

    def _calculate_derivatives(self) -> np.ndarray:
        """Calculate derivatives at each point using finite differences."""
        dydx = np.zeros_like(self._y)
        n = len(self._x)

        # For the first point, use forward difference
        dydx[0] = (self._y[1] - self._y[0]) / (self._x[1] - self._x[0])

        # For the last point, use backward difference
        dydx[-1] = (self._y[-1] - self._y[-2]) / (self._x[-1] - self._x[-2])

        # For the interior points, use central difference
        for i in range(1, n - 1):
            dydx[i] = (self._y[i + 1] - self._y[i - 1]) / (
                self._x[i + 1] - self._x[i - 1]
            )

        return dydx

    def cubic_hermite_interpolation(self, return_func=False) -> np.ndarray:
        self._dydx = self._calculate_derivatives()
        func_no_extrap = scipy.interpolate.CubicHermiteSpline(
            self._x, self._y, self._dydx 
        )
        if return_func:
            return func_no_extrap
        
        ynew_no_extrap = func_no_extrap(self._linspace_x)
        ynew = ynew_no_extrap.copy()
        if self._enable_extrapolate_left_fill:
            left_extrap_values = (
                self._dydx[0]
                * (self._linspace_x[self._linspace_x < self._x[0]] - self._x[0])
                + self._y[0]
            )
            ynew[self._linspace_x < self._x[0]] = left_extrap_values
        if self._enable_extrapolate_right_fill:
            right_extrap_values = (
                self._dydx[-1]
                * (self._linspace_x[self._linspace_x > self._x[-1]] - self._x[-1])
                + self._y[-1]
            )
            ynew[self._linspace_x > self._x[-1]] = right_extrap_values

        return ynew

    # monotonic cubic interpolation
    def pchip_interpolation(self, return_func=False) -> np.ndarray:
        func_no_extrap = scipy.interpolate.PchipInterpolator(
            self._x, self._y, extrapolate=False
        )
        func_extrap = scipy.interpolate.PchipInterpolator(
            self._x, self._y, extrapolate=True
        )
        if return_func:
            return func_no_extrap
        
        ynew_no_extrap = func_no_extrap(self._linspace_x)
        ynew_extrap = func_extrap(self._linspace_x)
        ynew = ynew_no_extrap.copy()
        if self._enable_extrapolate_left_fill:
            ynew[self._linspace_x < self._x[0]] = ynew_extrap[
                self._linspace_x < self._x[0]
            ]
        if self._enable_extrapolate_right_fill:
            ynew[self._linspace_x > self._x[-1]] = ynew_extrap[
                self._linspace_x > self._x[-1]
            ]

        return ynew

    def akima_interpolation(self, return_func=False) -> np.ndarray:
        akima = scipy.interpolate.Akima1DInterpolator(self._x, self._y)
        if return_func:
            return akima
        
        ynew = akima(self._linspace_x)
        if self._enable_extrapolate_left_fill:
            left_indices = self._linspace_x < min(self._x)
            ynew[left_indices] = self._y[0] + (
                self._linspace_x[left_indices] - self._x[0]
            ) * (self._y[1] - self._y[0]) / (self._x[1] - self._x[0])

        if self._enable_extrapolate_right_fill:
            right_indices = self._linspace_x > max(self._x)
            ynew[right_indices] = self._y[-1] + (
                self._linspace_x[right_indices] - self._x[-1]
            ) * (self._y[-1] - self._y[-2]) / (self._x[-1] - self._x[-2])

        return ynew

    def set_b_spline_k(self, b: int):
        self._bspline_k = b

    def b_spline1_interpolation(
        self, k: Optional[int] = None, return_func=False
    ) -> np.ndarray:
        if k:
            self._bspline_k = k
        spline = scipy.interpolate.make_interp_spline(
            self._x, self._y, k=self._bspline_k
        )
        if return_func:
            return spline
        
        ynew = spline(self._linspace_x)
        if self._enable_extrapolate_left_fill:
            left_indices = self._linspace_x < self._x[0]
            slope_left = (self._y[1] - self._y[0]) / (self._x[1] - self._x[0])
            ynew[left_indices] = self._y[0] + slope_left * (
                self._linspace_x[left_indices] - self._x[0]
            )

        if self._enable_extrapolate_right_fill:
            right_indices = self._linspace_x > self._x[-1]
            slope_right = (self._y[-1] - self._y[-2]) / (self._x[-1] - self._x[-2])
            ynew[right_indices] = self._y[-1] + slope_right * (
                self._linspace_x[right_indices] - self._x[-1]
            )

        return ynew

    def b_spline_with_knots_interpolation(
        self, knots: Optional[npt.ArrayLike], k: Optional[int] = 3, return_func=False
    ) -> np.ndarray:
        tck = scipy.interpolate.splrep(self._x, self._y, t=knots, k=k)
        # bspline = splev(ttm_interp, tck)
        bspline = scipy.interpolate.BSpline(*tck)
        if return_func:
            return bspline

        ynew_no_extrap = bspline(self._linspace_x)
        ynew_extrap = bspline(self._linspace_x)
        ynew = ynew_no_extrap.copy()
        if self._enable_extrapolate_left_fill:
            ynew[self._linspace_x < self._x[0]] = ynew_extrap[
                self._linspace_x < self._x[0]
            ]
        if self._enable_extrapolate_right_fill:
            ynew[self._linspace_x > self._x[-1]] = ynew_extrap[
                self._linspace_x > self._x[-1]
            ]

        return ynew

    def univariate_spline(self, s: float, return_func=False):
        unispline = scipy.interpolate.UnivariateSpline(
            x=self._x,
            y=self._y,
            s=s
        )
        if return_func:
            return unispline
        
        ynew_no_extrap = unispline(self._linspace_x)
        ynew_extrap = unispline(self._linspace_x)
        ynew = ynew_no_extrap.copy()
        if self._enable_extrapolate_left_fill:
            ynew[self._linspace_x < self._x[0]] = ynew_extrap[
                self._linspace_x < self._x[0]
            ]
        if self._enable_extrapolate_right_fill:
            ynew[self._linspace_x > self._x[-1]] = ynew_extrap[
                self._linspace_x > self._x[-1]
            ]

        return ynew

    
    def ppoly_interpolation(self, return_func=False) -> np.ndarray:
        coeffs = scipy.interpolate.CubicSpline(self._x, self._y).c
        ppoly = scipy.interpolate.PPoly(coeffs, self._x)
        if return_func:
            return ppoly

        ynew_no_extrap = ppoly(self._linspace_x)
        ynew = ynew_no_extrap.copy()
        if self._enable_extrapolate_left_fill:
            left_extrap_values = ppoly(self._linspace_x[self._linspace_x < self._x[0]])
            ynew[self._linspace_x < self._x[0]] = left_extrap_values
        if self._enable_extrapolate_right_fill:
            right_extrap_values = ppoly(
                self._linspace_x[self._linspace_x > self._x[-1]]
            )
            ynew[self._linspace_x > self._x[-1]] = right_extrap_values

        return ynew

    def monotone_convex(self) -> np.ndarray:
        mc_spline = MonotoneConvex(terms=self._x, spots=self._y)
        def mc_spline_func(t):
            return [mc_spline.spot(val) for val in t]
        return mc_spline_func
    
    def smoothing_spline(self, w=None, lam=None):
        return scipy.interpolate.make_smoothing_spline(self._x, self._y, w=w, lam=lam)
    
    def lsq_univariate_soline(self, knots: np.ndarray, k=3):
        return scipy.interpolate.LSQUnivariateSpline(self._x, self._y, t=knots, k=k)

    def plotter(
        self,
        linear: Optional[bool] = False,
        log_linear: Optional[bool] = False,
        cubic: Optional[bool] = False,
        cubic_hermite: Optional[bool] = False,
        cubic_bc_types_n_knots_n_nu: Optional[
            List[
                Tuple[
                    Literal["not-a-knot", "periodic", "clamped", "natural"],
                    np.ndarray,
                    int,
                ]
            ]
        ] = None,
        pchip: Optional[bool] = False,
        akima: Optional[bool] = False,
        b_spline: Optional[bool] = False,
        ppoly: Optional[bool] = False,
        run_parallel: Optional[bool] = False,
    ):
        def plot_helper(ynew: np.ndarray, title: str):
            plt.figure(figsize=(17, 6))
            plt.plot(self._linspace_x, ynew, label="Interpolated")
            plt.plot(self._x, self._y, "o", label="CUSIPs")
            plt.xlabel("Maturity")
            plt.ylabel("Yield")
            plt.title(title)
            plt.legend()
            plt.show()

        plottables: Dict[Callable[[], np.ndarray], str] = {
            self._linear_interpolation: "Linear Interpolation" if linear else None,
            self._log_linear_interpolation: (
                "Log-Linear Interpolation" if log_linear else None
            ),
            self._cubic_spline_interpolation: (
                f"Cubic Spline Interpolation - {self._cubic_spline_interp_bc_type}"
                if cubic
                else None
            ),
            self._cubic_hermite_interpolation: (
                f"Cubic Hermite Spline Interpolation" if cubic_hermite else None
            ),
            self._pchip_interpolation: (
                "PCHIP (Monotonic & Hermite Cubic) Interpolation" if pchip else None
            ),
            self._akima_interpolation: "Akima Interpolation" if akima else None,
            self._b_spline_interpolation: (
                f"B-Spline Interpolation - k = {self._bspline_k}" if b_spline else None
            ),
            self._ppoly_interpolation: (f"PPoly Interpolation" if ppoly else None),
        }

        if run_parallel:
            to_plot = {func: title for func, title in plottables.items() if title}
            with ThreadPoolExecutor() as executor:
                future_to_title = {
                    executor.submit(func): title for func, title in to_plot.items()
                }
                for future in as_completed(future_to_title):
                    title = future_to_title[future]
                    try:
                        ynew = future.result()
                        plot_helper(ynew, title)
                    except Exception as exc:
                        print(f"{title} generated an exception: {exc}")
        else:
            for func, title in plottables.items():
                if title:
                    ynew = func()
                    plot_helper(ynew, title)

        if cubic_bc_types_n_knots_n_nu:
            for bc_type, knots, nu in cubic_bc_types_n_knots_n_nu:
                if not knots:
                    knots = np.array([])
                if not nu:
                    nu = 0
                self.set_cubic_spline_interpolation_bc_type(bc_type=bc_type)
                self.set_cubic_spline_interpolation_custom_knots(custom_knots=knots)
                self.set_cubic_spline_interpolation_nu(nu=nu)
                ynew = self._cubic_spline_interpolation()
                title = f"Cubic Spline Interpolation - Boundary condition type: {self._cubic_spline_interp_bc_type} - Knots: {self._cubic_spline_interp_custom_knots}"
                plot_helper(ynew, title)
